#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.callsign_lookup import *


class TestCallsignLookup(unittest.TestCase):

    """ The unit tests for the callsign lookup functionality. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """
        self.qrz = CallsignLookupQRZ(parent=None)
        self.hamqth = CallsignLookupHamQTH(parent=None)

    def test_strip(self):
        """ Check that a callsign with a prefix and a suffix is stripped correctly. """
        assert strip("F/MYCALL/MM") == "MYCALL"
        assert strip("F/MYCALL/M") == "MYCALL"
        assert strip("HB9/MYCALL/P") == "MYCALL"
        assert strip("HB9/MYCALL/QRP") == "MYCALL"
        assert strip("HB9/MYCALL/A") == "MYCALL"
        assert strip("HB9/MYCALL/AM") == "MYCALL"
        assert strip("HB9/MYCALL/PM") == "MYCALL"

    def test_strip_prefix_only(self):
        """ Check that a callsign with only a prefix is stripped correctly. """
        assert strip("F/MYCALL") == "MYCALL"

    def test_strip_suffix_only(self):
        """ Check that a callsign with only a suffix is stripped correctly. """
        assert strip("MYCALL/M") == "MYCALL"
        assert strip("MYCALL/P") == "MYCALL"
        assert strip("MYCALL/A") == "MYCALL"
        assert strip("MYCALL/MM") == "MYCALL"
        assert strip("MYCALL/PM") == "MYCALL"
        assert strip("MYCALL/AM") == "MYCALL"
        assert strip("MYCALL/QRP") == "MYCALL"

    def test_strip_no_prefix_or_suffix(self):
        """ Check that a callsign with no prefix or suffix remains unmodified. """
        callsign = "MYCALL"
        assert strip(callsign) == "MYCALL"

    def test_strip_too_many_components(self):
        """ Check that a callsign with too many prefix/suffix components remains unmodified. """
        callsign = "F/HB9/MYCALL/MM"
        assert strip(callsign) == "F/HB9/MYCALL/MM"

    def test_qrz_connect(self):
        """ Check the example response from the qrz.com server, and make sure the session key has been correctly extracted. """

        http_client.HTTPConnection = mock.Mock(spec=http_client.HTTPConnection)
        http_client.HTTPResponse = mock.Mock(spec=http_client.HTTPResponse)
        connection = http_client.HTTPConnection()
        response = http_client.HTTPResponse()

        response.read.return_value = b'<?xml version="1.0" encoding="utf-8" ?>\n<QRZDatabase version="1.33" xmlns="http://xmldata.qrz.com">\n<Session>\n<Key>3b1fd1d3ba495189984f93ff67bd45b6</Key>\n<Count>61</Count>\n<SubExp>non-subscriber</SubExp>\n<GMTime>Sun Nov 22 21:25:34 2015</GMTime>\n<Remark>cpu: 0.147s</Remark>\n</Session>\n</QRZDatabase>\n'
        connection.getresponse.return_value = response

        result = self.qrz.connect("hello", "world")
        assert(result)
        assert(self.qrz.session_key == "3b1fd1d3ba495189984f93ff67bd45b6")

    def test_qrz_lookup(self):
        """ Check the example callsign lookup response from the qrz.com server, and make sure the callsign information has been correctly extracted. """

        http_client.HTTPConnection = mock.Mock(spec=http_client.HTTPConnection)
        http_client.HTTPResponse = mock.Mock(spec=http_client.HTTPResponse)
        connection = http_client.HTTPConnection()
        response = http_client.HTTPResponse()

        response.read.return_value = b'<?xml version="1.0" encoding="utf-8" ?>\n<QRZDatabase version="1.33" xmlns="http://xmldata.qrz.com">\n<Callsign>\n<call>MYCALL</call>\n<fname>FIRSTNAME</fname>\n<name>LASTNAME</name>\n<addr2>ADDRESS2</addr2>\n<country>COUNTRY</country>\n</Callsign>\n<Session>\n<Key>3b1fd1d3ba495189984f93ff67bd45b6</Key>\n<Count>61</Count>\n<SubExp>non-subscriber</SubExp>\n<Message>A subscription is required to access the complete record.</Message>\n<GMTime>Sun Nov 22 21:34:46 2015</GMTime>\n<Remark>cpu: 0.026s</Remark>\n</Session>\n</QRZDatabase>\n'
        connection.getresponse.return_value = response

        self.qrz.connection = connection
        self.qrz.session_key = "3b1fd1d3ba495189984f93ff67bd45b6"
        fields_and_data = self.qrz.lookup("MYCALL")
        assert(fields_and_data["NAME"] == "FIRSTNAME LASTNAME")
        assert(fields_and_data["ADDRESS"] == "ADDRESS2")
        assert(fields_and_data["COUNTRY"] == "COUNTRY")

    def test_hamqth_connect(self):
        """ Check the example response from the hamqth.com server, and make sure the session ID has been correctly extracted. """

        http_client.HTTPSConnection = mock.Mock(spec=http_client.HTTPSConnection)
        http_client.HTTPResponse = mock.Mock(spec=http_client.HTTPResponse)
        connection = http_client.HTTPSConnection()
        response = http_client.HTTPResponse()

        response.read.return_value = b'<?xml version="1.0"?>\n<HamQTH version="2.6" xmlns="https://www.hamqth.com">\n<session>\n<session_id>09b0ae90050be03c452ad235a1f2915ad684393c</session_id>\n</session>\n</HamQTH>\n'
        connection.getresponse.return_value = response

        result = self.hamqth.connect("hello", "world")
        assert(result)
        assert(self.hamqth.session_id == "09b0ae90050be03c452ad235a1f2915ad684393c")

    def test_hamqth_lookup(self):
        """ Check the example callsign lookup response from the hamqth.com server, and make sure the callsign information has been correctly extracted. """

        http_client.HTTPSConnection = mock.Mock(spec=http_client.HTTPSConnection)
        http_client.HTTPResponse = mock.Mock(spec=http_client.HTTPResponse)
        connection = http_client.HTTPSConnection()
        response = http_client.HTTPResponse()

        response.read.return_value = b'<?xml version="1.0"?>\n<HamQTH version="2.6" xmlns="https://www.hamqth.com">\n<search>\n<callsign>MYCALL</callsign>\n<nick>NAME</nick>\n<country>COUNTRY</country>\n<itu>ITU</itu>\n<cq>CQ</cq>\n<iota>IOTA</iota>\n<adr_street1>ADDRESS</adr_street1>\n</search>\n</HamQTH>\n'
        connection.getresponse.return_value = response

        self.hamqth.connection = connection
        self.hamqth.session_id = "09b0ae90050be03c452ad235a1f2915ad684393c"
        fields_and_data = self.hamqth.lookup("MYCALL")
        assert(fields_and_data["NAME"] == "NAME")
        assert(fields_and_data["ADDRESS"] == "ADDRESS")
        assert(fields_and_data["COUNTRY"] == "COUNTRY")
        assert(fields_and_data["CQZ"] == "CQ")
        assert(fields_and_data["ITUZ"] == "ITU")
        assert(fields_and_data["IOTA"] == "IOTA")

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


class Blank(object):

    """ A blank page in the logbook for the "+" (New Log) tab. """

    def __init__(self, application):
        """ Create the blank page.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application

        blank_treeview = Gtk.TreeView()

        # Allow the (blank) page to be scrolled up/down
        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(blank_treeview)
        page = Gtk.VBox()
        page.pack_start(sw, True, True, 0)

        # Add a "+" button to the tab
        tab = Gtk.HBox(homogeneous=False, spacing=0)
        icon = Gtk.Image.new_from_icon_name(Gtk.STOCK_ADD, Gtk.IconSize.MENU)
        button = Gtk.Button()
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.set_focus_on_click(False)
        button.connect("clicked", self.application.logbook.new_log)
        button.add(icon)
        button.set_tooltip_text('New Log')
        tab.pack_start(button, False, False, 0)

        tab.show_all()
        page.show_all()

        self.application.logbook.notebook.insert_page(page, tab, 1)
        self.application.logbook.notebook.set_current_page(0)

        return



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.summary import *
from pyqso.logbook import Logbook


class TestSummary(unittest.TestCase):

    """ The unit tests for the Summary class. """

    def setUp(self):
        """ Set up the objects needed for the unit tests and create a connection to the test database. """
        PyQSO = mock.MagicMock()
        self.summary = Summary(application=PyQSO())
        self.summary.logbook = Logbook(application=PyQSO())
        path_to_test_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "test.db")
        success = self.summary.logbook.db_connect(path_to_test_database)
        assert(success)
        self.summary.logbook.logs = self.summary.logbook.get_logs()
        assert(self.summary.logbook.logs is not None)

    def tearDown(self):
        """ Destroy the connection to the test database. """
        success = self.summary.logbook.db_disconnect()
        assert(success)

    def test_get_year_bounds(self):
        """ Check that the years of the earliest and latest QSO are correct. """
        year_min, year_max = self.summary.get_year_bounds()
        assert(year_min == 2012)
        assert(year_max == 2017)

    def test_get_annual_contact_count(self):
        """ Check that there are 3 QSOs in the year 2017. """
        count = self.summary.get_annual_contact_count(2017)
        april = datetime(2017, 4, 1)
        april_count = count[april]
        assert(april_count == 3)  # A total of 3 contacts made in April.
        assert(sum(count.values()) == 4)  # A total of 4 contacts made that whole year.

    def test_get_annual_mode_count(self):
        """ Check that, in the year 2017, there was 1 QSO made using CW, 2 QSOs made using FM, and 1 QSO made using SSB. """
        count = self.summary.get_annual_mode_count(2017)
        assert(count["CW"] == 1)
        assert(count["FM"] == 2)
        assert(count["SSB"] == 1)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from pyqso.cabrillo import *


class TestCabrillo(unittest.TestCase):

    """ The unit tests for the Cabrillo class. """

    def setUp(self):
        """ Set up the Cabrillo object needed for the unit tests. """
        self.cabrillo = Cabrillo()
        return

    def test_write(self):
        """ Check that QSOs are written correctly in Cabrillo format. """
        records = [{'TIME_ON': '1955', 'BAND': '40m', 'CALL': 'TEST', 'FREQ': "145.550", 'MODE': 'FM', 'QSO_DATE': '20130322', 'RST_SENT': '59 001', 'RST_RCVD': '59 002'}, {'TIME_ON': '0820', 'BAND': '20m', 'CALL': 'TEST2ABC', 'FREQ': "144.330", 'MODE': 'SSB', 'QSO_DATE': '20150227', 'RST_SENT': '55 020', 'RST_RCVD': '57 003'}, {'TIME_ON': '0832', 'BAND': '2m', 'CALL': 'HELLO', 'FREQ': "145.550", 'MODE': 'FM', 'QSO_DATE': '20150227', 'RST_SENT': '59 001', 'RST_RCVD': '59 002'}]

        expected = """START-OF-LOG: 3.0
CREATED-BY: PyQSO v1.1.0
CALLSIGN: MYCALL
CONTEST: MYCONTEST
QSO: 145550.0 FM 2013-03-22 1955 MYCALL 59 001 TEST 59 002 0
QSO: 144330.0 PH 2015-02-27 0820 MYCALL 55 020 TEST2ABC 57 003 0
QSO: 145550.0 FM 2015-02-27 0832 MYCALL 59 001 HELLO 59 002 0
END-OF-LOG:"""
        print("Expected Cabrillo file contents: ", expected)

        mycall = "MYCALL"
        mycontest = "MYCONTEST"
        path = "Cabrillo.test_write.log"
        self.cabrillo.write(records, path, contest=mycontest, mycall=mycall)

        actual = ""
        f = open(path, "r")
        for line in f:
            actual += line
        f.close()
        print("Actual Cabrillo file contents: ", actual)
        assert(expected == actual)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import logging

CABRILLO_VERSION = "3.0"

CONTESTS = ["", "AP-SPRINT", "ARRL-10", "ARRL-160", "ARRL-222", "ARRL-DX-CW", "ARRL-DX-SSB", "ARRL-RR-PH", "ARRL-RR-DIG", "ARRL-RR-CW", "ARRL-SCR", "ARRL-SS-CW", "ARRL-SS-SSB", "ARRL-UHF-AUG", "ARRL-VHF-JAN", "ARRL-VHF-JUN", "ARRL-VHF-SEP", "ARRL-RTTY", "BARTG-RTTY", "CQ-160-CW", "CQ-160-SSB", "CQ-WPX-CW", "CQ-WPX-RTTY", "CQ-WPX-SSB", "CQ-VHF", "CQ-WW-CW", "CQ-WW-RTTY", "CQ-WW-SSB", "DARC-WAEDC-CW", "DARC-WAEDC-RTTY", "DARC-WAEDC-SSB", "DL-DX-RTTY", "DRCG-WW-RTTY", "FCG-FQP", "IARU-HF", "JIDX-CW", "JIDX-SSB", "NAQP-CW", "NAQP-SSB", "NA-SPRINT-CW", "NA-SPRINT-SSB", "NCCC-CQP", "NEQP", "OCEANIA-DX-CW", "OCEANIA-DX-SSB", "RDXC", "RSGB-IOTA", "SAC-CW", "SAC-SSB", "STEW-PERRY", "TARA-RTTY"]


class Cabrillo:

    """ The Cabrillo class supplies methods for writing log files in the Cabrillo format (v3.0).
    For more information, visit http://wwrof.org/cabrillo/ """

    def __init__(self):
        """ Initialise class for I/O of files using the Cabrillo format. """
        return

    def write(self, records, path, contest="", mycall=""):
        """ Write a list of QSO records to a file in the Cabrillo format.

        :arg list records: The list of QSO records to write.
        :arg str path: The desired path of the Cabrillo file to write to.
        :arg str contest: The name of the contest.
        :arg str mycall: The callsign used during the contest.
        :returns: None
        :raises IOError: If the Cabrillo file cannot be written (e.g. due to lack of write permissions)."""

        logging.debug("Writing records to a Cabrillo file...")

        with open(path, mode='w', errors="replace") as f:  # Open file for writing

            # Header
            f.write("""START-OF-LOG: %s\n""" % (CABRILLO_VERSION))
            f.write("""CREATED-BY: PyQSO v1.1.0\n""")
            f.write("""CALLSIGN: %s\n""" % (mycall))
            f.write("""CONTEST: %s\n""" % (contest))

            # Write each record to the file.
            for r in records:

                # Frequency. Note that this must be in kHz. The frequency is stored in MHz in the database, so it's converted to kHz here.
                try:
                    freq = str(float(r["FREQ"])*1e3)
                except ValueError:
                    freq = ""

                # Mode
                if(r["MODE"] == "SSB"):
                    mo = "PH"
                elif(r["MODE"] == "CW"):
                    mo = "CW"
                elif(r["MODE"] == "FM"):
                    mo = "FM"
                else:
                    # FIXME: This assumes that the mode is any other non-CW digital mode, which isn't always going to be the case (e.g. for AM).
                    mo = "RY"

                # Date in yyyy-mm-dd format.
                date = r["QSO_DATE"][0:4] + "-" + r["QSO_DATE"][4:6] + "-" + r["QSO_DATE"][6:8]

                # Time
                time = r["TIME_ON"]

                # The callsign that was used when operating the contest station.
                call_sent = mycall

                # Exchange (the part sent to the distant station)
                exch_sent = r["RST_SENT"]

                # Callsign
                call_rcvd = r["CALL"]

                # Exchange (the part received from the distant station)
                exch_rcvd = r["RST_RCVD"]

                # Transmitter ID (must be 0 or 1, if applicable).
                # FIXME: For now this has been hard-coded to 0.
                t = "0"

                f.write("""QSO: %s %s %s %s %s %s %s %s %s\n""" % (freq, mo, date, time, call_sent, exch_sent, call_rcvd, exch_rcvd, t))

            # Footer
            f.write("END-OF-LOG:")

            logging.info("Wrote %d QSOs to %s in Cabrillo format." % (len(records), path))

        return



#!/usr/bin/env python3

#    Copyright (C) 2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.


class Popup:

    """ The popup menu that appears when a QSO record is right-clicked. """

    def __init__(self, application):
        """ Set up popup menu items. """

        self.application = application
        self.builder = self.application.builder

        self.menu = self.builder.get_object("popup")

        # Collect Gtk menu items and connect signals.
        self.items = {}

        # Plot selected QSO on the world map.
        self.items["PINPOINT"] = self.builder.get_object("mitem_pinpoint")
        self.items["PINPOINT"].connect("activate", self.application.logbook.pinpoint_callback)

        self.items["COPY"] = self.builder.get_object("mitem_copy")
        self.items["COPY"].connect("activate", self.application.logbook.copy_callback)

        self.items["PASTE"] = self.builder.get_object("mitem_paste")
        self.items["PASTE"].connect("activate", self.application.logbook.paste_callback)

        return



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.calendar_dialog import *


class TestCalendarDialog(unittest.TestCase):

    """ The unit tests for the CalendarDialog class. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """
        self.cd = CalendarDialog(application=mock.MagicMock())
        self.cd.calendar = Gtk.Calendar()
        self.cd.calendar.select_month(3, 2017)  # Note: Months start from 0 when using the Calendar widget. So "3" represents April here.
        self.cd.calendar.select_day(2)

    def test_date(self):
        """ Check that the date obtained from the Calendar is in the correct format. """
        assert(self.cd.date == "20170402")

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import os


class TelnetConnectionDialog:

    """ A handler for the Gtk.Dialog through which a user can specify Telnet connection details. """

    def __init__(self, application):
        """ Create and show the Telnet connection dialog to the user.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.builder = application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("telnet_connection_dialog",))
        self.dialog = self.builder.get_object("telnet_connection_dialog")
        self.sources = {"HOST": self.builder.get_object("host_entry"),
                        "PORT": self.builder.get_object("port_entry"),
                        "USERNAME": self.builder.get_object("username_entry"),
                        "PASSWORD": self.builder.get_object("password_entry"),
                        "BOOKMARK": self.builder.get_object("bookmark_checkbox")}

        self.dialog.show_all()

        return

    @property
    def host(self):
        """ Return the Telnet server's host name.

        :returns: The server's host name.
        :rtype: str
        """
        return self.sources["HOST"].get_text()

    @property
    def port(self):
        """ Return the Telnet server's port number (as a string).

        :returns: The server's port number (as a string).
        :rtype: str
        """
        return self.sources["PORT"].get_text()

    @property
    def username(self):
        """ Return the user's username.

        :returns: The user's username.
        :rtype: str
        """
        return self.sources["USERNAME"].get_text()

    @property
    def password(self):
        """ Return the user's password.

        :returns: The user's password.
        :rtype: str
        """
        return self.sources["PASSWORD"].get_text()

    @property
    def bookmark(self):
        """ Return True if a new bookmark should be created, otherwise return False.

        :returns: True if a new bookmark should be created, otherwise False.
        :rtype: bool
        """
        return self.sources["BOOKMARK"].get_active()



#!/usr/bin/env python3

#    Copyright (C) 2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.world_map import Maidenhead, WorldMap


class TestMaidenhead(unittest.TestCase):

    """ The unit tests for the Maidenhead class. """

    def setUp(self):
        """ Set up the Maidenhead object needed for the unit tests. """
        self.maidenhead = Maidenhead()

    def test_ll2gs(self):
        """ Check that a latitude-longitude coordinate can correctly be converted to a Maidenhead grid square. """
        latitude = 51.0593
        longitude = -1.4262
        assert self.maidenhead.ll2gs(latitude, longitude, subsquare=False) == "IO91"
        assert self.maidenhead.ll2gs(latitude, longitude, subsquare=True) == "IO91gb"

    def test_gs2ll(self):
        """ Check that a Maidenhead grid square can correctly be converted to a latitude-longitude coordinate. """
        gs4 = "JN05"
        assert self.maidenhead.gs2ll(gs4) == (45.5, 1.0)
        gs6 = "JN05aa"
        assert self.maidenhead.gs2ll(gs6) == (45.020833333333336, 0.041666666666666664)
        gs6 = "IO91gb"
        assert self.maidenhead.gs2ll(gs6) == (51.0625, -1.4583333333333335)


class TestWorldMap(unittest.TestCase):

    """ The unit tests for the WorldMap class. """

    def setUp(self):
        """ Set up the WorldMap object needed for the unit tests. """
        PyQSO = mock.MagicMock()
        self.world_map = WorldMap(application=PyQSO())

    def test_get_worked_grid_squares(self):
        """ Check that the worked grid squares are determined correctly. """
        Logbook = mock.MagicMock()
        Log = mock.MagicMock()
        logbook = Logbook()
        l = Log()
        l.records = [{"CALL": "TEST123", "COUNTRY": "England", "GRIDSQUARE": "IO91gb"}, {"CALL": "TEST456", "COUNTRY": "England", "GRIDSQUARE": "IO90hv"}, {"CALL": "TEST789", "COUNTRY": "England", "GRIDSQUARE": None}]
        logbook.logs = [l]
        worked_grid_squares = self.world_map.get_worked_grid_squares(logbook=logbook)
        assert worked_grid_squares[14, 8]  # IO square.

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.awards import *
from pyqso.logbook import Logbook


class TestAwards(unittest.TestCase):

    """ The unit tests for the Awards class. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """
        PyQSO = mock.MagicMock()
        self.awards = Awards(application=PyQSO())
        self.logbook = Logbook(application=PyQSO())
        path_to_test_database = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "test.db")
        success = self.logbook.db_connect(path_to_test_database)
        assert(success)
        self.logbook.logs = self.logbook.get_logs()
        assert(self.logbook.logs is not None)

    def test_count(self):
        """ Check that there are 3 FM/AM/SSB/SSTV QSOs and 1 CW QSO. Note that the BAND must be specified in order to be counted. """
        count = self.awards.count(self.logbook)
        assert(sum(count[0]) == 3)  # FM/AM/SSB/SSTV
        assert(sum(count[1]) == 1)  # CW
        assert(sum(count[2]) == 1)  # Other modes
        assert(sum(count[3]) == 5)  # Mixed

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path


class LogNameDialog:

    """ A handler for the Gtk.Dialog through which a user can specify the name of a Log object. """

    def __init__(self, application, title=None, name=None):
        """ Create and show the log name dialog to the user.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        :arg title: The title of the dialog. If this is None, it is assumed that a new log is going to be created.
        :arg name: The existing name of the Log object. Defaults to None if not specified (because the Log does not yet exist).
        """

        logging.debug("Building new log name dialog...")

        self.builder = application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("log_name_dialog",))
        self.dialog = self.builder.get_object("log_name_dialog")

        if(title is None):
            self.dialog.set_title("New Log")
        else:
            self.dialog.set_title(title)

        self.entry = self.builder.get_object("log_name_entry")
        if(name is not None):
            self.entry.set_text(name)

        self.dialog.show_all()

        logging.debug("Log name dialog built.")

        return

    @property
    def name(self):
        """ Return the log name specified in the Gtk.Entry box by the user.

        :returns: The log's name.
        :rtype: str
        """
        return self.entry.get_text()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyqso.dx_cluster import *


class TestDXCluster(unittest.TestCase):

    """ The unit tests for the DXCluster class. """

    def setUp(self):
        """ Set up the objects needed for the unit tests. """

        PyQSO = mock.MagicMock()
        self.dxcluster = DXCluster(application=PyQSO())

    def test_on_telnet_io(self):
        """ Check that the response from the Telnet server can be correctly decoded. """

        telnetlib.Telnet = mock.Mock(spec=telnetlib.Telnet)
        connection = telnetlib.Telnet("hello", "world")
        connection.read_very_eager.return_value = b"Test message from the Telnet server."
        self.dxcluster.connection = connection
        result = self.dxcluster.on_telnet_io()
        assert(result)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from pyqso.log import *


class TestLog(unittest.TestCase):

    """ The unit tests for the Log class. """

    def setUp(self):
        """ Create a connection to a temporary database and set up the objects needed for the unit tests. """
        self.connection = sqlite.connect(":memory:")
        self.connection.row_factory = sqlite.Row

        self.field_names = ["CALL", "QSO_DATE", "TIME_ON", "FREQ", "BAND", "MODE", "RST_SENT", "RST_RCVD"]
        self.fields_and_data = {"CALL": "TEST123", "QSO_DATE": "20130312", "TIME_ON": "1234", "FREQ": "145.500", "BAND": "2m", "MODE": "FM", "RST_SENT": "59", "RST_RCVD": "59"}

        c = self.connection.cursor()
        query = "CREATE TABLE test (id INTEGER PRIMARY KEY AUTOINCREMENT"
        for field_name in self.field_names:
            s = ", %s TEXT" % field_name.lower()
            query = query + s
        query = query + ")"
        c.execute(query)

        self.log = Log(self.connection, "test")

    def tearDown(self):
        """ Destroy the connection to the temporary database. """
        self.connection.close()

    def test_add_missing_db_columns(self):
        """ Check that any missing columns in the database are added successfully. """

        c = self.connection.cursor()

        # 'Before' state.
        column_names_before = []
        c.execute("PRAGMA table_info(test)")
        result = c.fetchall()
        for t in result:
            column_names_before.append(t[1].upper())

        # Add missing columns.
        self.log.add_missing_db_columns()

        # 'After' state.
        column_names_after = []
        c.execute("PRAGMA table_info(test)")
        result = c.fetchall()
        for t in result:
            column_names_after.append(t[1].upper())

        print("Column names before: ", column_names_before)
        print("Column names after: ", column_names_after)

        assert(len(column_names_before) == len(self.field_names) + 1)  # Added 1 here because of the "id" column in all database tables.
        assert(len(column_names_after) == len(AVAILABLE_FIELD_NAMES_ORDERED) + 1)
        for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
            assert(field_name in column_names_after)

    def test_add_record(self):
        """ Check that a single record can be successfully added. """

        self.log.add_record(self.fields_and_data)

        c = self.connection.cursor()
        c.execute("SELECT * FROM test")
        records = c.fetchall()
        assert len(records) == 1

        # Check that all the data has been added to all the fields.
        for field_name in self.field_names:
            print(self.fields_and_data[field_name], records[0][field_name])
            assert self.fields_and_data[field_name] == records[0][field_name]

        # Check consistency of index between Gtk.ListStore and the database.
        assert(records[0]["id"] == 1)
        iter = self.log.get_iter_first()
        row_index = self.log.get_value(iter, 0)
        assert(records[0]["id"] == row_index)

    def test_add_record_multiple(self):
        """ Check that multiple records can be successfully added in one go. """
        self.log.add_record([self.fields_and_data]*5)

        c = self.connection.cursor()
        c.execute("SELECT * FROM test")
        records = c.fetchall()
        assert len(records) == 5

    def test_delete_record(self):
        """ Check that a record can be successfully deleted. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))

        c.execute("SELECT * FROM test")
        records_before = c.fetchall()

        self.log.delete_record(1)

        c.execute("SELECT * FROM test")
        records_after = c.fetchall()

        assert(len(records_before) == 1)
        assert(len(records_after) == 0)

    def test_edit_record(self):
        """ Check that a record's fields can be successfully edited. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))

        c.execute("SELECT * FROM test")
        record_before = c.fetchall()[0]

        self.log.edit_record(1, "CALL", "TEST456")
        self.log.edit_record(1, "FREQ", "145.450")

        c.execute("SELECT * FROM test")
        record_after = c.fetchall()[0]

        assert(record_before["CALL"] == "TEST123")
        assert(record_after["CALL"] == "TEST456")
        assert(record_before["FREQ"] == "145.500")
        assert(record_after["FREQ"] == "145.450")

    def test_get_record_by_index(self):
        """ Check that a record can be retrieved using its index. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))

        record = self.log.get_record_by_index(1)
        print("Contents of retrieved record: ", record)
        for field_name in list(record.keys()):
            if(field_name.upper() == "ID"):
                continue
            else:
                assert(record[field_name.upper()] == self.fields_and_data[field_name.upper()])
        assert(len(record) == len(self.fields_and_data) + 1)

    def test_records(self):
        """ Check that all records in a log can be successfully retrieved. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        # Add the same record twice
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))

        records = self.log.records
        print("Contents of all retrieved records: ", records)
        assert(len(records) == 2)  # There should be 2 records
        for field_name in self.field_names:
            assert(records[0][field_name] == self.fields_and_data[field_name])
            assert(records[1][field_name] == self.fields_and_data[field_name])

    def test_record_count(self):
        """ Check that the total number of records in a log is calculated correctly. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        # Add the same record twice
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))
        c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))

        record_count = self.log.record_count
        print("Number of records in the log: ", record_count)
        assert(record_count == 2)  # There should be 2 records

    def test_get_duplicates(self):
        """ Insert n records, n-1 of which are duplicates, and check that the duplicates are successfully identified. """
        query = "INSERT INTO test VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?)"
        c = self.connection.cursor()
        n = 5  # The total number of records to insert.
        for i in range(0, n):
            c.execute(query, (self.fields_and_data["CALL"], self.fields_and_data["QSO_DATE"], self.fields_and_data["TIME_ON"], self.fields_and_data["FREQ"], self.fields_and_data["BAND"], self.fields_and_data["MODE"], self.fields_and_data["RST_SENT"], self.fields_and_data["RST_RCVD"]))
        assert(len(self.log.get_duplicates()) == n-1)  # Expecting n-1 duplicates.

    def test_remove_duplicates(self):
        """ Insert n records, n-1 of which are duplicates, and check that the duplicates are successfully removed. """
        n = 5  # The total number of records to insert.
        for i in range(0, n):
            self.log.add_record(self.fields_and_data)
        (number_of_duplicates, number_of_duplicates_removed) = self.log.remove_duplicates()
        print("Number of duplicates: %d" % number_of_duplicates)
        print("Number of duplicates removed: %d" % number_of_duplicates_removed)
        assert(number_of_duplicates == number_of_duplicates_removed)
        assert(number_of_duplicates == 4)
        assert(self.log.record_count == 1)

    def test_rename(self):
        """ Check that a log can be successfully renamed. """
        old_name = "test"
        new_name = "hello"
        success = self.log.rename(new_name)
        assert(success)
        with self.connection:
            c = self.connection.cursor()
            c.execute("SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE name=?)", [old_name])
            exists = c.fetchone()
            assert(exists[0] == 0)  # Old log name should no longer exist.
            c.execute("SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE name=?)", [new_name])
            exists = c.fetchone()
            assert(exists[0] == 1)  # New log name should now exist.
        assert(self.log.name == new_name)

if(__name__ == '__main__'):
    unittest.main()



#!/usr/bin/env python3

#    Copyright (C) 2012-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import os.path


class Menu:

    """ The menu bar along the top of the main window. """

    def __init__(self, application):
        """ Set up all menu items and connect to the various functions.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application
        self.builder = self.application.builder

        # Collect Gtk menu items and connect signals.
        self.items = {}

        # New logbook
        self.items["NEW_LOGBOOK"] = self.builder.get_object("mitem_new_logbook")
        self.items["NEW_LOGBOOK"].connect("activate", self.application.logbook.new)

        # Open logbook
        self.items["OPEN_LOGBOOK"] = self.builder.get_object("mitem_open_logbook")
        self.items["OPEN_LOGBOOK"].connect("activate", self.application.logbook.open)

        # Close logbook
        self.items["CLOSE_LOGBOOK"] = self.builder.get_object("mitem_close_logbook")
        self.items["CLOSE_LOGBOOK"].connect("activate", self.application.logbook.close)

        # New log
        self.items["NEW_LOG"] = self.builder.get_object("mitem_new_log")
        self.items["NEW_LOG"].connect("activate", self.application.logbook.new_log)

        # Delete the current log
        self.items["DELETE_LOG"] = self.builder.get_object("mitem_delete_log")
        self.items["DELETE_LOG"].connect("activate", self.application.logbook.delete_log)

        # Rename the current log
        self.items["RENAME_LOG"] = self.builder.get_object("mitem_rename_log")
        self.items["RENAME_LOG"].connect("activate", self.application.logbook.rename_log)

        # Import log
        self.items["IMPORT_LOG"] = self.builder.get_object("mitem_import_log")
        self.items["IMPORT_LOG"].connect("activate", self.application.logbook.import_log)

        # Export the current log as ADIF
        self.items["EXPORT_LOG_ADIF"] = self.builder.get_object("mitem_export_log_adif")
        self.items["EXPORT_LOG_ADIF"].connect("activate", self.application.logbook.export_log_adif)

        # Export the current log as Cabrillo
        self.items["EXPORT_LOG_CABRILLO"] = self.builder.get_object("mitem_export_log_cabrillo")
        self.items["EXPORT_LOG_CABRILLO"].connect("activate", self.application.logbook.export_log_cabrillo)

        # Print log
        self.items["PRINT_LOG"] = self.builder.get_object("mitem_print_log")
        self.items["PRINT_LOG"].connect("activate", self.application.logbook.print_log)

        # Preferences
        self.items["PREFERENCES"] = self.builder.get_object("mitem_preferences")
        self.items["PREFERENCES"].connect("activate", self.application.show_preferences)

        # Quit
        self.items["QUIT"] = self.builder.get_object("mitem_quit")
        self.items["QUIT"].connect("activate", Gtk.main_quit)

        # Add record
        self.items["ADD_RECORD"] = self.builder.get_object("mitem_add_record")
        self.items["ADD_RECORD"].connect("activate", self.application.logbook.add_record_callback)

        # Edit selected record
        self.items["EDIT_RECORD"] = self.builder.get_object("mitem_edit_record")
        self.items["EDIT_RECORD"].connect("activate", self.application.logbook.edit_record_callback)

        # Delete selected record
        self.items["DELETE_RECORD"] = self.builder.get_object("mitem_delete_record")
        self.items["DELETE_RECORD"].connect("activate", self.application.logbook.delete_record_callback)

        # Remove duplicates
        self.items["REMOVE_DUPLICATES"] = self.builder.get_object("mitem_remove_duplicates")
        self.items["REMOVE_DUPLICATES"].connect("activate", self.application.logbook.remove_duplicates_callback)

        # Record count
        self.items["RECORD_COUNT"] = self.builder.get_object("mitem_record_count")
        self.items["RECORD_COUNT"].connect("activate", self.application.logbook.record_count_callback)

        # View toolbox
        self.items["TOOLBOX"] = self.builder.get_object("mitem_toolbox")
        config = configparser.ConfigParser()
        have_config = (config.read(os.path.expanduser('~/.config/pyqso/preferences.ini')) != [])
        (section, option) = ("general", "show_toolbox")
        if(have_config and config.has_option(section, option)):
            self.items["TOOLBOX"].set_active(config.getboolean(section, option))
        else:
            self.items["TOOLBOX"].set_active(False)  # Don't show the toolbox by default
        self.items["TOOLBOX"].connect("activate", self.application.toolbox.toggle_visible_callback)

        # About
        self.items["ABOUT"] = self.builder.get_object("mitem_about")
        self.items["ABOUT"].connect("activate", self.application.show_about)

        self.set_logbook_item_sensitive(True)
        self.set_log_items_sensitive(False)
        self.set_record_items_sensitive(False)

        return

    def set_logbook_item_sensitive(self, sensitive):
        """ Enable/disable logbook-related menu items.

        :arg bool sensitive: If True, enable the 'new logbook' and 'open logbook' menu items. If False, disable them.
        """

        self.items["NEW_LOGBOOK"].set_sensitive(sensitive)
        self.items["OPEN_LOGBOOK"].set_sensitive(sensitive)
        self.items["CLOSE_LOGBOOK"].set_sensitive(not sensitive)
        return

    def set_log_items_sensitive(self, sensitive):
        """ Enable/disable log-related menu items.

        :arg bool sensitive: If True, enable all the log-related menu items. If False, disable them all.
        """

        for item_name in ["NEW_LOG", "DELETE_LOG", "RENAME_LOG", "IMPORT_LOG", "EXPORT_LOG_ADIF", "EXPORT_LOG_CABRILLO", "PRINT_LOG"]:
            self.items[item_name].set_sensitive(sensitive)
        return

    def set_record_items_sensitive(self, sensitive):
        """ Enable/disable record-related menu items.

        :arg bool sensitive: If True, enable all the record-related menu items. If False, disable them all.
        """

        for item_name in ["ADD_RECORD", "EDIT_RECORD", "DELETE_RECORD", "REMOVE_DUPLICATES", "RECORD_COUNT"]:
            self.items[item_name].set_sensitive(sensitive)
        return



#!/usr/bin/env python3

#    Copyright (C) 2013-2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from pyqso.dx_cluster import DXCluster
from pyqso.world_map import WorldMap
from pyqso.awards import Awards


class Toolbox:

    """ Contains a Gtk.Notebook full of amateur radio-related tools. """

    def __init__(self, application):
        """ Instantiate and insert the various tools into the toolbox.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application
        self.builder = self.application.builder

        self.tools = self.builder.get_object("tools")

        self.dx_cluster = DXCluster(self.application)
        self.world_map = WorldMap(self.application)
        self.awards = Awards(self.application)

        self.tools.connect_after("switch-page", self.on_switch_page)

        return

    def toggle_visible_callback(self, widget=None):
        """ Show/hide the toolbox. """
        toolbox_frame = self.builder.get_object("toolbox")
        toolbox_frame.set_visible(not toolbox_frame.get_visible())
        return

    def on_switch_page(self, widget, label, new_page):
        """ Re-draw the WorldMap if the user switches to the World Map tab. """
        if(widget.get_tab_label(label).get_text() == "World Map"):
            self.world_map.draw()
        return



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
from pyqso.cabrillo import CONTESTS


class CabrilloExportDialog:

    """ A handler for the Gtk.Dialog through which a user can specify Cabrillo log details. """

    def __init__(self, application):
        """ Create and show the Cabrillo export dialog to the user.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        logging.debug("Building new Cabrillo export dialog...")

        self.builder = application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("cabrillo_export_dialog",))
        self.dialog = self.builder.get_object("cabrillo_export_dialog")

        self.contest_combo = self.builder.get_object("cabrillo_export_contest_combo")
        self.mycall_entry = self.builder.get_object("cabrillo_export_mycall_entry")
        for contest in CONTESTS:
            self.contest_combo.append_text(contest)

        self.dialog.show_all()

        logging.debug("Cabrillo export dialog built.")

        return

    @property
    def contest(self):
        """ Return the name of the contest.

        :returns: The name of the contest.
        :rtype: str
        """
        return self.contest_combo.get_active_text()

    @property
    def mycall(self):
        """ Return the callsign used during the contest.

        :returns: The callsign used during the contest.
        :rtype: str
        """
        # Always show the callsigns in upper case.
        return self.mycall_entry.get_text().upper()



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path


class CalendarDialog:

    """ Handler for a simple dialog containing a Gtk.Calendar widget. Using this ensures the date is in the correct YYYYMMDD format required by ADIF. """

    def __init__(self, application):
        """ Set up the calendar widget and show it to the user.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.builder = application.builder
        glade_file_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "res", "pyqso.glade")
        self.builder.add_objects_from_file(glade_file_path, ("calendar_dialog",))
        self.dialog = self.builder.get_object("calendar_dialog")
        self.calendar = self.builder.get_object("calendar")
        self.dialog.show_all()

        return

    @property
    def date(self):
        """ Return the date from the Gtk.Calendar widget in YYYYMMDD format.

        :returns: The date from the calendar in YYYYMMDD format.
        :rtype: str
        """
        logging.debug("Retrieving the date from the calendar...")
        (year, month, day) = self.calendar.get_date()
        # If necessary, add on leading zeros so the YYYYMMDD format is followed.
        if(month + 1 < 10):
            month = "0" + str(month + 1)  # Note: the months start from an index of 0 when retrieved from the calendar widget.
        else:
            month += 1
        if(day < 10):
            day = "0" + str(day)
        date = str(year) + str(month) + str(day)
        return date



#!/usr/bin/env python3

#    Copyright (C) 2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, PangoCairo
import logging

from pyqso.auxiliary_dialogs import error


class Printer(object):

    """ Handles the printing of one or more records to file or paper. """

    def __init__(self, application):
        """ Initialise the record printer.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application

        self.action = Gtk.PrintOperationAction.PRINT_DIALOG
        self.operation = Gtk.PrintOperation()
        ps = Gtk.PageSetup()
        ps.set_orientation(Gtk.PageOrientation.LANDSCAPE)
        self.operation.set_default_page_setup(ps)
        self.operation.set_unit(Gtk.Unit.MM)

        self.operation.connect("begin_print", self.begin_print)
        self.operation.connect("draw_page", self.draw_page)

        return

    def print_records(self, records, title=None):
        """ Perform the print operation.

        :arg dict records: The records to be printed.
        :arg str title: Optional title for the document. Default is None.
        :returns: The result of the print operation.
        :rtype: Gtk.PrintOperationResult
        """

        # Add the title, if given.
        if(title):
            self.text_to_print = title + "\n\n"
        else:
            self.text_to_print = ""

        # Assemble the header and records into one string.
        line_format = "%-5s\t%-15s\t%-8s\t%-6s\t%-15s\t%-12s\t%-8s\t%-8s\n"
        self.text_to_print += line_format % ("Index", "Callsign", "Date", "Time", "Frequency", "Mode", "RST Sent", "RST Rcvd")
        self.text_to_print += line_format % ("-----", "--------", "----", "----", "---------", "----", "--------", "--------")
        for r in records:
            self.text_to_print += line_format % (str(r["id"]), str(r["CALL"]), str(r["QSO_DATE"]), str(r["TIME_ON"]), str(r["FREQ"]), str(r["MODE"]), str(r["RST_SENT"]), str(r["RST_RCVD"]))

        result = self.operation.run(self.action, parent=self.application.window)
        if(result == Gtk.PrintOperationResult.ERROR):
            error(parent=self.application.window, message="Unable to print the log.")

        return result

    def begin_print(self, operation, context):
        """ Specify the layout/position/font of the text on the pages to be printed.

        :arg Gtk.PrintOperation operation: The printing API.
        :arg Gtk.PrintContext context: Used to draw/render the pages to print.
        """
        width = context.get_width()  # Measured in pixels.
        height = context.get_height()  # Measured in pixels.
        layout = context.create_pango_layout()
        layout.set_font_description(Pango.FontDescription("monospace expanded 10"))
        layout.set_width(int(width*Pango.SCALE))
        layout.set_text(self.text_to_print, -1)

        number_of_pages = 1
        page_height = 0
        for line in range(0, layout.get_line_count()):
            layout_line = layout.get_line(line)
            ink_rectangle, logical_rectangle = layout_line.get_pixel_extents()
            self.line_height = logical_rectangle.height + 3.0
            page_height += self.line_height
            if((page_height + 2*self.line_height) >= height):
                # Go on to the next page.
                number_of_pages += 1
                page_height = 0.0
        operation.set_n_pages(number_of_pages)
        logging.debug("Printing %d pages..." % number_of_pages)
        self.text_to_print = self.text_to_print.split("\n")
        return

    def draw_page(self, operation, context, page_number):
        """ Render the QSO details on the page.

        :arg Gtk.PrintOperation operation: The printing API.
        :arg Gtk.PrintContext context: Used to draw/render the pages to print.
        :arg int page_number: The current page number.
        """
        cr = context.get_cairo_context()
        cr.set_source_rgb(0, 0, 0)
        layout = context.create_pango_layout()
        layout.set_font_description(Pango.FontDescription("monospace expanded 10"))
        layout.set_width(int(context.get_width()*Pango.SCALE))

        current_line_number = 1
        for line in self.text_to_print:
            layout.set_text(line, -1)
            cr.move_to(5, current_line_number*self.line_height)
            PangoCairo.update_layout(cr, layout)
            PangoCairo.show_layout(cr, layout)
            current_line_number += 1
            if((current_line_number+1)*self.line_height >= context.get_height()):
                for j in range(0, current_line_number-1):
                    self.text_to_print.pop(0)  # Remove what has been printed already before draw_page is called again.
                break

        return



#!/usr/bin/env python3

#    Copyright (C) 2013-2017 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

import logging
try:
    import http.client as http_client
except ImportError:
    import httplib as http_client
from xml.dom import minidom
try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from pyqso.auxiliary_dialogs import error


class CallsignLookupQRZ:

    """ Use qrz.com to lookup details about a particular callsign. """

    def __init__(self, parent):
        """ Initialise a new callsign lookup handler.

        :arg parent: The parent Gtk dialog.
        """
        self.parent = parent
        self.connection = None
        self.session_key = None
        return

    def connect(self, username, password):
        """ Initiate a session with the qrz.com server. Hopefully this will provide a session key.

        :arg str username: The username of the qrz.com user account.
        :arg str password: The password of the qrz.com user account.
        :returns: True if a successful connection was made to the server, and False otherwise.
        :rtype: bool
        """
        logging.debug("Connecting to the qrz.com server...")

        # Connect to the server.
        try:
            self.connection = http_client.HTTPConnection("xmldata.qrz.com")
            request = "/xml/current/?username=%s;password=%s;agent=pyqso" % (username, quote(password))  # Percent-escape the password in case there are reserved characters present.
            self.connection.request("GET", request)
            response = self.connection.getresponse()
        except Exception as e:
            logging.exception(e)
            error(parent=self.parent, message="Could not connect to the qrz.com server. Check connection to the internets?")
            return False

        # Get the session key.
        xml_data = minidom.parseString(response.read())
        session_node = xml_data.getElementsByTagName("Session")[0]  # There should only be one Session element.
        session_key_node = session_node.getElementsByTagName("Key")
        if(session_key_node):
            self.session_key = session_key_node[0].firstChild.nodeValue
            logging.debug("Successfully connected to the qrz.com server. Session key is: %s." % self.session_key)
            connected = True
        else:
            connected = False

        # If there are any errors or warnings, print them out.
        session_error_node = session_node.getElementsByTagName("Error")
        if(session_error_node):
            session_error = session_error_node[0].firstChild.nodeValue
            error(parent=self.parent, message="qrz.com session error: %s" % session_error)

        return connected

    def lookup(self, full_callsign, ignore_prefix_suffix=True):
        """ Parse the XML tree that is returned from the qrz.com XML server to obtain the NAME, ADDRESS, STATE, COUNTRY, DXCC, CQZ, ITUZ, and IOTA field data (if present).

        :arg str full_callsign: The callsign to look up (without any prefix/suffix stripping).
        :arg bool ignore_prefix_suffix: True if callsign prefixes/suffixes should be removed prior to querying the server, False otherwise.
        :returns: The data in a dictionary called fields_and_data.
        :rtype: dict
        """

        logging.debug("Looking up callsign. The full callsign (with a prefix and/or suffix) is %s." % full_callsign)

        # Remove any prefix or suffix from the callsign before performing the lookup.
        if(ignore_prefix_suffix):
            callsign = strip(full_callsign)
        else:
            callsign = full_callsign

        # Commence lookup.
        fields_and_data = {"NAME": "", "ADDRESS": "", "STATE": "", "COUNTRY": "", "DXCC": "", "CQZ": "", "ITUZ": "", "IOTA": ""}
        if(self.session_key):
            request = "/xml/current/?s=%s;callsign=%s" % (self.session_key, callsign)
            self.connection.request("GET", request)
            response = self.connection.getresponse()

            xml_data = minidom.parseString(response.read())
            callsign_node = xml_data.getElementsByTagName("Callsign")
            if(callsign_node):
                callsign_node = callsign_node[0]  # There should only be a maximum of one Callsign element.

                callsign_fname_node = callsign_node.getElementsByTagName("fname")
                callsign_name_node = callsign_node.getElementsByTagName("name")
                if(callsign_fname_node):
                    fields_and_data["NAME"] = callsign_fname_node[0].firstChild.nodeValue
                if(callsign_name_node):  # Add the surname, if present.
                    fields_and_data["NAME"] = fields_and_data["NAME"] + " " + callsign_name_node[0].firstChild.nodeValue

                callsign_addr1_node = callsign_node.getElementsByTagName("addr1")
                callsign_addr2_node = callsign_node.getElementsByTagName("addr2")
                if(callsign_addr1_node):
                    fields_and_data["ADDRESS"] = callsign_addr1_node[0].firstChild.nodeValue
                if(callsign_addr2_node):  # Add the second line of the address, if present.
                    fields_and_data["ADDRESS"] = (fields_and_data["ADDRESS"] + ", " if callsign_addr1_node else "") + callsign_addr2_node[0].firstChild.nodeValue

                callsign_state_node = callsign_node.getElementsByTagName("state")
                if(callsign_state_node):
                    fields_and_data["STATE"] = callsign_state_node[0].firstChild.nodeValue

                callsign_country_node = callsign_node.getElementsByTagName("country")
                if(callsign_country_node):
                    fields_and_data["COUNTRY"] = callsign_country_node[0].firstChild.nodeValue

                callsign_ccode_node = callsign_node.getElementsByTagName("ccode")
                if(callsign_ccode_node):
                    fields_and_data["DXCC"] = callsign_ccode_node[0].firstChild.nodeValue

                callsign_cqzone_node = callsign_node.getElementsByTagName("cqzone")
                if(callsign_cqzone_node):
                    fields_and_data["CQZ"] = callsign_cqzone_node[0].firstChild.nodeValue

                callsign_ituzone_node = callsign_node.getElementsByTagName("ituzone")
                if(callsign_ituzone_node):
                    fields_and_data["ITUZ"] = callsign_ituzone_node[0].firstChild.nodeValue

                callsign_iota_node = callsign_node.getElementsByTagName("iota")
                if(callsign_iota_node):
                    fields_and_data["IOTA"] = callsign_iota_node[0].firstChild.nodeValue
            else:
                # If there is no Callsign element, then print out the error message in the Session element.
                session_node = xml_data.getElementsByTagName("Session")
                if(session_node):
                    session_error_node = session_node[0].getElementsByTagName("Error")
                    if(session_error_node):
                        session_error = session_error_node[0].firstChild.nodeValue
                        error(parent=self.parent, message=session_error)
                # Return empty strings for the field data.
            logging.debug("Callsign lookup complete. Returning data...")
        return fields_and_data


class CallsignLookupHamQTH:

    """ Use hamqth.com to lookup details about a particular callsign. """

    def __init__(self, parent):
        self.parent = parent
        self.connection = None
        self.session_id = None
        return

    def connect(self, username, password):
        """ Initiate a session with the hamqth.com server. Hopefully this will provide a session key.

        :arg str username: The username of the hamqth.com user account.
        :arg str password: The password of the hamqth.com user account.
        :returns: True if a successful connection was made to the server, and False otherwise.
        :rtype: bool
        """

        logging.debug("Connecting to the hamqth.com server...")

        # Connect to the server.
        try:
            self.connection = http_client.HTTPSConnection("www.hamqth.com")
            request = "/xml.php?u=%s&p=%s" % (username, quote(password))  # Percent-escape the password in case there are reserved characters present.
            self.connection.request("GET", request)
            response = self.connection.getresponse()
        except Exception as e:
            logging.exception(e)
            error(parent=self.parent, message="Could not connect to the hamqth.com server. Check connection to the internets?")
            return False

        # Get the session ID.
        xml_data = minidom.parseString(response.read())
        session_node = xml_data.getElementsByTagName("session")[0]  # There should only be one Session element.
        session_id_node = session_node.getElementsByTagName("session_id")
        if(session_id_node):
            self.session_id = session_id_node[0].firstChild.nodeValue
            logging.debug("Successfully connected to the hamqth.com server. Session ID is: %s." % self.session_id)
            connected = True
        else:
            connected = False

        # If there are any errors or warnings, print them out.
        session_error_node = session_node.getElementsByTagName("error")
        if(session_error_node):
            session_error = session_error_node[0].firstChild.nodeValue
            error(parent=self.parent, message="hamqth.com session error: %s" % session_error)

        return connected

    def lookup(self, full_callsign, ignore_prefix_suffix=True):
        """ Parse the XML tree that is returned from the hamqth.com XML server to obtain the NAME, ADDRESS, STATE, COUNTRY, DXCC, CQZ, ITUZ, and IOTA field data (if present),

        :arg str full_callsign: The callsign to look up (without any prefix/suffix stripping).
        :arg bool ignore_prefix_suffix: True if callsign prefixes/suffixes should be removed prior to querying the server, False otherwise.
        :returns: The data in a dictionary called fields_and_data.
        :rtype: dict
        """

        logging.debug("Looking up callsign. The full callsign (with a prefix and/or suffix) is %s." % full_callsign)

        # Remove any prefix or suffix from the callsign before performing the lookup.
        if(ignore_prefix_suffix):
            callsign = strip(full_callsign)
        else:
            callsign = full_callsign

        # Commence lookup.
        fields_and_data = {"NAME": "", "ADDRESS": "", "STATE": "", "COUNTRY": "", "DXCC": "", "CQZ": "", "ITUZ": "", "IOTA": ""}
        if(self.session_id):
            request = "/xml.php?id=%s&callsign=%s&prg=pyqso" % (self.session_id, callsign)
            self.connection.request("GET", request)
            response = self.connection.getresponse()

            xml_data = minidom.parseString(response.read())
            search_node = xml_data.getElementsByTagName("search")
            if(search_node):
                search_node = search_node[0]  # There should only be a maximum of one Callsign element.

                search_name_node = search_node.getElementsByTagName("nick")
                if(search_name_node):
                    fields_and_data["NAME"] = search_name_node[0].firstChild.nodeValue

                search_addr1_node = search_node.getElementsByTagName("adr_street1")
                search_addr2_node = search_node.getElementsByTagName("adr_street2")
                if(search_addr1_node):
                    fields_and_data["ADDRESS"] = search_addr1_node[0].firstChild.nodeValue
                if(search_addr2_node):  # Add the second line of the address, if present.
                    fields_and_data["ADDRESS"] = (fields_and_data["ADDRESS"] + ", " if search_addr1_node else "") + search_addr2_node[0].firstChild.nodeValue

                search_state_node = search_node.getElementsByTagName("us_state")
                if(search_state_node):
                    fields_and_data["STATE"] = search_state_node[0].firstChild.nodeValue

                search_country_node = search_node.getElementsByTagName("country")
                if(search_country_node):
                    fields_and_data["COUNTRY"] = search_country_node[0].firstChild.nodeValue

                search_cqzone_node = search_node.getElementsByTagName("cq")
                if(search_cqzone_node):
                    fields_and_data["CQZ"] = search_cqzone_node[0].firstChild.nodeValue

                search_ituzone_node = search_node.getElementsByTagName("itu")
                if(search_ituzone_node):
                    fields_and_data["ITUZ"] = search_ituzone_node[0].firstChild.nodeValue

                search_iota_node = search_node.getElementsByTagName("iota")
                if(search_iota_node):
                    fields_and_data["IOTA"] = search_iota_node[0].firstChild.nodeValue
            else:
                # If there is no Callsign element, then print out the error message in the Session element.
                session_node = xml_data.getElementsByTagName("session")
                if(session_node):
                    session_error_node = session_node[0].getElementsByTagName("error")
                    if(session_error_node):
                        session_error = session_error_node[0].firstChild.nodeValue
                        error(parent=self.parent, message=session_error)
                # Return empty strings for the field data.

            logging.debug("Callsign lookup complete. Returning data...")
        return fields_and_data


def strip(full_callsign):
    """ Remove any prefixes or suffixes from a callsign.

    :arg str full_callsign: The callsign to be considered for prefix/suffix removal.
    :returns: The callsign with prefixes/suffixes removed.
    :rtype: str
    """

    components = full_callsign.split("/")  # We assume that prefixes or suffixes come before/after a forward slash character "/".
    suffixes = ["P", "M", "A", "PM", "MM", "AM", "QRP"]
    try:
        if(len(components) == 3):
            # We have both a prefix and a suffix.
            callsign = components[1]

        elif(len(components) == 2):
            if(components[1].upper() in suffixes or components[1].lower() in suffixes):
                # If the last part of the full_callsign is a valid suffix, then use the part before that.
                callsign = components[0]
                logging.debug("Suffix %s found. Callsign to lookup is %s." % (components[1], callsign))
            else:
                # We have a prefix, so take the part after the first "/".
                callsign = components[1]
                logging.debug("Prefix %s found. Callsign to lookup is %s." % (components[0], callsign))

        elif(len(components) == 1):
            # We have neither a prefix nor a suffix, so use the full_callsign.
            callsign = full_callsign
            logging.debug("No prefix or suffix found. Callsign to lookup is %s." % callsign)

        else:
            raise ValueError
    except ValueError:
        callsign = full_callsign
    return callsign



#!/usr/bin/env python3

#    Copyright (C) 2012-2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
import logging
import sqlite3 as sqlite
import json
from os.path import expanduser
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from pyqso.adif import *
from pyqso.cabrillo import *
from pyqso.log import *
from pyqso.auxiliary_dialogs import *
from pyqso.log_name_dialog import LogNameDialog
from pyqso.record_dialog import RecordDialog
from pyqso.cabrillo_export_dialog import CabrilloExportDialog
from pyqso.summary import Summary
from pyqso.blank import Blank
from pyqso.printer import Printer
from pyqso.compare import compare_date_and_time, compare_default


class Logbook:

    """ A Logbook object can store multiple Log objects. """

    def __init__(self, application):
        """ Create a new Logbook object and initialise the list of Logs.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """

        self.application = application
        self.builder = self.application.builder
        self.notebook = self.builder.get_object("logbook")
        self.connection = None
        self.logs = []

        return

    def new(self, widget=None):
        """ Create a new logbook, and open it.

        :returns: True if the new logbook is successfully opened, and False otherwise.
        :rtype: bool
        """

        # Get the new file's path from a dialog.
        dialog = Gtk.FileChooserDialog("Create a New SQLite Database File",
                                       self.application.window,
                                       Gtk.FileChooserAction.SAVE,
                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
        dialog.set_do_overwrite_confirmation(True)

        response = dialog.run()
        if(response == Gtk.ResponseType.OK):
            path = dialog.get_filename()
        else:
            path = None
        dialog.destroy()

        if(path is None):  # If the Cancel button has been clicked, path will still be None.
            logging.debug("No file path specified.")
            return
        else:
            # Clear the contents of the file, in case the file exists already.
            open(path, 'w').close()
            # Open the new logbook, ready for use.
            opened = self.open(path=path)
        return opened

    def open(self, widget=None, path=None):
        """ Open a logbook, and render all the logs within it.

        :arg str path: An optional argument containing the database file location, if already known. If this is None, a file selection dialog will appear.
        :returns: True if the logbook is successfully opened, and False otherwise.
        :rtype: bool
        """

        if(path is None):
            # If no path has been provided, get one from a "File Open" dialog.
            dialog = Gtk.FileChooserDialog("Open SQLite Database File",
                                           self.application.window,
                                           Gtk.FileChooserAction.OPEN,
                                           (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                            Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

            response = dialog.run()
            if(response == Gtk.ResponseType.OK):
                path = dialog.get_filename()
            dialog.destroy()

            if(path is None):  # If the Cancel button has been clicked, path will still be None.
                logging.debug("No file path specified.")
                return False

        connected = self.db_connect(path)
        if(connected):
            # If the connection setup was successful, then open all the logs in the database.

            self.path = path

            logging.debug("Retrieving all the logs in the logbook...")
            try:
                self.logs = self.get_logs()
            except (sqlite.Error, IndexError) as e:
                logging.exception(e)
                error(parent=self.application.window, message="Could not open logbook. Something went wrong when trying to retrieve the logs. Perhaps the logbook file is encrypted, corrupted, or in the wrong format?")
                return False

            logging.debug("All logs retrieved successfully.")

            logging.debug("Rendering logs...")
            # For rendering the logs. One treeview and one treeselection per Log.
            self.treeview = []
            self.treeselection = []
            self.sorter = []
            self.filter = []
            self.summary = Summary(self.application)
            self.blank = Blank(self.application)

            # FIXME: This is an unfortunate work-around. If the area around the "+/New Log" button
            # is clicked, PyQSO will change to an empty page. This signal is used to stop this from happening.
            self.notebook.connect("switch-page", self.on_switch_page)

            for i in range(len(self.logs)):
                self.render_log(i)
            logging.debug("All logs rendered successfully.")

            self.summary.update()
            self.application.toolbox.awards.count(self)

            context_id = self.application.statusbar.get_context_id("Status")
            self.application.statusbar.push(context_id, "Logbook: %s" % self.path)
            self.application.toolbar.set_logbook_button_sensitive(False)
            self.application.menu.set_logbook_item_sensitive(False)
            self.application.menu.set_log_items_sensitive(True)
            self.application.toolbar.filter_source.set_sensitive(True)

            self.notebook.show_all()

        else:
            logging.debug("Not connected to a logbook. No logs were opened.")
            return False

        return True

    def close(self, widget=None):
        """ Close the logbook that is currently open.

        :returns: True if the logbook is successfully closed, and False otherwise.
        :rtype: bool
        """

        disconnected = self.db_disconnect()
        if(disconnected):
            logging.debug("Closing all logs in the logbook...")
            while(self.notebook.get_n_pages() > 0):
                # Once a page is removed, the other pages get re-numbered,
                # so a 'for' loop isn't the best option here.
                self.notebook.remove_page(0)
            logging.debug("All logs now closed.")

            context_id = self.application.statusbar.get_context_id("Status")
            self.application.statusbar.push(context_id, "No logbook is currently open.")
            self.application.toolbar.set_logbook_button_sensitive(True)
            self.application.menu.set_logbook_item_sensitive(True)
            self.application.menu.set_log_items_sensitive(False)
            self.application.toolbar.filter_source.set_sensitive(False)
        else:
            logging.debug("Unable to disconnect from the database. No logs were closed.")
            return False

        return True

    def db_connect(self, path):
        """ Create an SQL database connection to the Logbook's data source.

        :arg str path: The path of the database file.
        """

        logging.debug("Attempting to connect to the logbook database...")
        # Try setting up the SQL database connection.
        try:
            self.db_disconnect()  # Destroy any existing connections first.
            self.connection = sqlite.connect(path)
            self.connection.row_factory = sqlite.Row
        except sqlite.Error as e:
            # Cannot connect to the database.
            logging.exception(e)
            error(parent=self.application.window, message="Cannot connect to the database. Check file permissions?")
            return False

        logging.debug("Database connection created successfully!")
        return True

    def db_disconnect(self):
        """ Destroy the connection to the Logbook's data source.

        :returns: True if the connection was successfully destroyed, and False otherwise.
        :rtype: bool
        """

        logging.debug("Cleaning up any existing database connections...")
        if(self.connection):
            try:
                self.connection.close()
            except sqlite.Error as e:
                logging.exception(e)
                return False
        else:
            logging.debug("Already disconnected. Nothing to do here.")
        return True

    def on_switch_page(self, widget, label, new_page):
        """ Handle a tab/page change, and enable/disable the relevant Record-related buttons. """

        if(new_page == self.notebook.get_n_pages()-1):  # The last (right-most) tab is the "New Log" tab.
            self.notebook.stop_emission("switch-page")

        # Disable the record buttons if a log page is not selected.
        if(new_page == 0):
            self.application.toolbar.set_record_buttons_sensitive(False)
            self.application.menu.set_record_items_sensitive(False)
        else:
            self.application.toolbar.set_record_buttons_sensitive(True)
            self.application.menu.set_record_items_sensitive(True)
        return

    def on_button_release_event(self, treeview, event):
        """ Show a popup menu when the user right-clicks a record in the logbook. """

        if(event.button == 3):
            self.application.popup.menu.popup(None, None, None, None, event.button, event.time)
            self.application.popup.menu.show_all()
            return True

    def new_log(self, widget=None):
        """ Create a new log in the logbook. """

        if(self.connection is None):
            return
        exists = True
        ln = LogNameDialog(self.application)
        while(exists):
            response = ln.dialog.run()
            if(response == Gtk.ResponseType.OK):
                log_name = ln.name
                try:
                    with self.connection:
                        c = self.connection.cursor()
                        # NOTE: "id" is simply an alias for the "rowid" column here.
                        query = "CREATE TABLE %s (id INTEGER PRIMARY KEY AUTOINCREMENT" % log_name
                        for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
                            s = ", %s TEXT" % field_name.lower()
                            query = query + s
                        query = query + ")"
                        c.execute(query)
                        exists = False
                except sqlite.Error as e:
                    logging.exception(e)
                    # Data is not valid - inform the user.
                    error(parent=ln.dialog, message="Database error. Try another log name.")
                    exists = True
            else:
                ln.dialog.destroy()
                return

        ln.dialog.destroy()

        # Instantiate and populate a new Log object.
        l = Log(self.connection, log_name)
        l.populate()

        self.logs.append(l)
        self.render_log(self.log_count-1)
        self.summary.update()

        self.notebook.set_current_page(self.log_count)
        return

    def delete_log(self, widget, page=None):
        """ Delete the log that is currently selected in the logbook.

        :arg Gtk.Widget page: An optional argument corresponding to the currently-selected page/tab.
        """
        if(self.connection is None):
            return

        if(page is None):
            page_index = self.notebook.get_current_page()  # Get the index of the selected tab in the logbook.
            if(page_index == 0):  # If we are on the Summary page...
                logging.debug("No log currently selected!")
                return
            else:
                page = self.notebook.get_nth_page(page_index)  # Get the Gtk.VBox of the selected tab in the logbook.

        log_index = self.get_log_index(name=page.get_name())
        log = self.logs[log_index]

        # We also need the page's index in order to remove it using remove_page below.
        # This may not be the same as what get_current_page() returns.
        page_index = self.notebook.page_num(page)

        if(page_index == 0 or page_index == self.notebook.get_n_pages()-1):  # Only the "New Log" tab is present (i.e. no actual logs in the logbook).
            logging.debug("No logs to delete!")
            return

        response = question(parent=self.application.window, message="Are you sure you want to delete log %s?" % log.name)
        if(response == Gtk.ResponseType.YES):
            try:
                with self.connection:
                    c = self.connection.cursor()
                    c.execute("DROP TABLE %s" % log.name)
            except sqlite.Error as e:
                logging.exception(e)
                error(parent=self.application.window, message="Database error. Could not delete the log.")
                return

            self.logs.pop(log_index)
            # Remove the log from the renderers too.
            self.treeview.pop(log_index)
            self.treeselection.pop(log_index)
            self.sorter.pop(log_index)
            self.filter.pop(log_index)
            # And finally remove the tab in the Logbook.
            self.notebook.remove_page(page_index)

        self.summary.update()
        self.application.toolbox.awards.count(self)
        return

    def filter_logs(self, widget=None):
        """ Re-filter all the logs when the user-defined expression is changed. """
        for i in range(0, len(self.filter)):
            self.filter[i].refilter()
        return

    def filter_by_callsign(self, model, iter, data):
        """ Filter all the logs in the logbook by the callsign field, based on a user-defined expression.

        :arg Gtk.TreeModel model: The model used to filter the log data.
        :arg Gtk.TreeIter iter: A pointer to a particular row in the model.
        :arg data: The user-defined expression to filter by.
        :returns: True if a record matches the expression, or if there is nothing to filter. Otherwise, returns False.
        :rtype: bool
        """
        value = model.get_value(iter, 1)
        callsign = self.application.toolbar.filter_source.get_text()

        if(callsign is None or callsign == ""):
            # If there is nothing to filter with, then show all the records!
            return True
        else:
            # This should be case insensitive.
            # Also, we could use value[:][0:len(callsign))] if we wanted to match from the very start of each callsign.
            return callsign.upper() in value or callsign.lower() in value

    def render_log(self, index):
        """ Render a Log in the Gtk.Notebook.

        :arg int index: The index of the Log (in the list of Logs) to render.
        """
        self.filter.append(self.logs[index].filter_new(root=None))
        # Set the callsign column as the column we want to filter by.
        self.filter[index].set_visible_func(self.filter_by_callsign, data=None)
        self.sorter.append(Gtk.TreeModelSort(model=self.filter[index]))
        self.sorter[index].set_sort_column_id(0, Gtk.SortType.ASCENDING)

        self.treeview.append(Gtk.TreeView(model=self.sorter[index]))
        self.treeview[index].set_grid_lines(Gtk.TreeViewGridLines.BOTH)
        self.treeview[index].connect("row-activated", self.edit_record_callback)
        self.treeview[index].connect("button-release-event", self.on_button_release_event)
        self.treeselection.append(self.treeview[index].get_selection())
        self.treeselection[index].set_mode(Gtk.SelectionMode.SINGLE)

        # Allow the Log to be scrolled up/down.
        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(self.treeview[index])
        vbox = Gtk.VBox()
        vbox.set_name(self.logs[index].name)  # Set a name for the tab itself so we can match it up with the associated Log object later.
        vbox.pack_start(sw, True, True, 0)

        # Add a close button to the tab
        hbox = Gtk.HBox(homogeneous=False, spacing=0)
        label = Gtk.Label(label=self.logs[index].name)
        hbox.pack_start(label, False, False, 0)
        hbox.show_all()

        self.notebook.insert_page(vbox, hbox, index+1)  # Append the new log as a new tab.

        # The first column of the logbook will always be the unique record index.
        # Let's append this separately to the field names.
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Index", renderer, text=0)
        column.set_resizable(True)
        column.set_min_width(50)
        column.set_clickable(True)
        column.set_sort_order(Gtk.SortType.ASCENDING)
        column.set_sort_indicator(True)
        column.connect("clicked", self.sort_log, 0)
        self.treeview[index].append_column(column)

        # Set up column names for each selected field
        field_names = AVAILABLE_FIELD_NAMES_ORDERED
        for i in range(0, len(field_names)):
            renderer = Gtk.CellRendererText()

            # Keep each row to a single line.
            renderer.set_property("single-paragraph-mode", True)

            column = Gtk.TreeViewColumn(AVAILABLE_FIELD_NAMES_FRIENDLY[field_names[i]], renderer, text=i+1)
            column.set_resizable(True)
            column.set_min_width(50)
            column.set_clickable(True)

            # Special cases
            if(field_names[i] == "NOTES"):
                # Give the 'Notes' column some extra space, since this is likely to contain some long sentences ...
                column.set_min_width(300)
                # ... but not too much extra space ...
                column.set_max_width(600)
                # ... and don't let the column automatically re-size itself.
                column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)

            column.connect("clicked", self.sort_log, i+1)

            config = configparser.ConfigParser()
            have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])
            (section, option) = ("view", AVAILABLE_FIELD_NAMES_ORDERED[i].lower())
            if(have_config and config.has_option(section, option)):
                column.set_visible(config.getboolean(section, option))
            self.treeview[index].append_column(column)

        self.notebook.show_all()
        return

    def sort_log(self, widget, column_index):
        """ Sort the log (that is currently selected) with respect to a given field.

        :arg int column_index: The index of the column to sort by.
        """

        log_index = self.get_log_index()
        column = self.treeview[log_index].get_column(column_index)

        if(AVAILABLE_FIELD_NAMES_ORDERED[column_index-1] == "QSO_DATE"):
            # If the field being sorted is the QSO_DATE, then also sort by the TIME_ON field so we get the
            # correct chronological order.
            # Note: This assumes that the TIME_ON field is always immediately to the right of the QSO_DATE field.
            self.sorter[log_index].set_sort_func(column_index, compare_date_and_time, user_data=[column_index, column_index+1])
        else:
            self.sorter[log_index].set_sort_func(column_index, compare_default, user_data=column_index)

        # If we are operating on the currently-sorted column...
        if(self.sorter[log_index].get_sort_column_id()[0] == column_index):
            order = column.get_sort_order()
            # ...then check if we need to reverse the order of searching.
            if(order == Gtk.SortType.ASCENDING):
                self.sorter[log_index].set_sort_column_id(column_index, Gtk.SortType.DESCENDING)
                column.set_sort_order(Gtk.SortType.DESCENDING)
            else:
                self.sorter[log_index].set_sort_column_id(column_index, Gtk.SortType.ASCENDING)
                column.set_sort_order(Gtk.SortType.ASCENDING)
        else:
            # Otherwise, change to the new sorted column. Default to ASCENDING order.
            self.sorter[log_index].set_sort_column_id(column_index, Gtk.SortType.ASCENDING)
            column.set_sort_order(Gtk.SortType.ASCENDING)

            # Show an arrow pointing in the direction of the sorting.
            # (First we need to remove the arrow from the previously-sorted column.
            # Since we don't know which one that was, just remove the arrow from all columns
            # and start again. This only loops over a few dozen columns at most, so
            # hopefully it won't take too much time.)
            for i in range(0, len(AVAILABLE_FIELD_NAMES_ORDERED)):
                column = self.treeview[log_index].get_column(i)
                column.set_sort_indicator(False)
            column = self.treeview[log_index].get_column(column_index)
            column.set_sort_indicator(True)
        return

    def rename_log(self, widget=None):
        """ Rename the log that is currently selected. """
        if(self.connection is None):
            return
        page_index = self.notebook.get_current_page()
        if(page_index == 0):  # If we are on the Summary page...
            logging.debug("No log currently selected!")
            return
        page = self.notebook.get_nth_page(page_index)  # Get the Gtk.VBox of the selected tab in the logbook.
        old_log_name = page.get_name()

        log_index = self.get_log_index(name=old_log_name)

        success = False
        ln = LogNameDialog(self.application, title="Rename Log", name=old_log_name)
        while(not success):
            response = ln.dialog.run()
            if(response == Gtk.ResponseType.OK):
                new_log_name = ln.name
                success = self.logs[log_index].rename(new_log_name)
                if(success):
                    ln.dialog.destroy()
                else:
                    # Unsuccessful rename attempt. Inform the user.
                    error(parent=ln.dialog, message="Database error. Try another log name.")
            else:
                ln.dialog.destroy()
                return

        # Remember to change the page's name ...
        page.set_name(new_log_name)

        # ... and update the tab's label.
        hbox = Gtk.HBox(homogeneous=False, spacing=0)
        label = Gtk.Label(label=new_log_name)
        hbox.pack_start(label, False, False, 0)
        hbox.show_all()
        self.notebook.set_tab_label(page, hbox)

        # The number of logs will obviously stay the same, but
        # we want to update the logbook's modification date.
        self.summary.update()

        return

    def import_log(self, widget=None):
        """ Import a log from an ADIF file. """

        # Get the path to the ADIF file.
        dialog = Gtk.FileChooserDialog("Import ADIF Log File",
                                       self.application.window,
                                       Gtk.FileChooserAction.OPEN,
                                      (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                       Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        filter = Gtk.FileFilter()
        filter.set_name("All ADIF files (*.adi, *.ADI)")
        filter.add_pattern("*.adi")
        filter.add_pattern("*.ADI")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if(response == Gtk.ResponseType.OK):
            path = dialog.get_filename()
        else:
            path = None
        dialog.destroy()

        if(path is None):
            logging.debug("No file path specified.")
            return

        # Read the records.
        adif = ADIF()
        try:
            records = adif.read(path)
        except IOError as e:
            error(parent=self.application.window, message="Could not import the log. I/O error %d: %s" % (e.errno, e.strerror))
            return
        except Exception as e:
            error(parent=self.application.window, message="Could not import the log.")
            logging.exception(e)
            return

        # Get the new log's name (or the name of the existing log the user wants to import into).
        ln = LogNameDialog(self.application, title="Import Log")
        while(True):
            response = ln.dialog.run()
            if(response == Gtk.ResponseType.OK):
                log_name = ln.name

                # Check if the log name exists.
                try:
                    exists = self.log_name_exists(log_name)
                except (sqlite.Error, IndexError) as e:
                    # Could not determine if the log name exists. It's safer to stop here than to try to add a new log.
                    logging.exception(e)
                    error(parent=ln.dialog, message="Database error. Could not check if the log name exists.")
                    ln.dialog.destroy()
                    return

                if(exists):
                    # Import into existing log.
                    l = self.logs[self.get_log_index(name=log_name)]
                    response = question(parent=ln.dialog, message="Are you sure you want to import into an existing log?")
                    if(response == Gtk.ResponseType.YES):
                        break
                else:
                    # Create a new log with the name the user supplies.
                    try:
                        with self.connection:
                            c = self.connection.cursor()
                            query = "CREATE TABLE %s (id INTEGER PRIMARY KEY AUTOINCREMENT" % log_name
                            for field_name in AVAILABLE_FIELD_NAMES_ORDERED:
                                s = ", %s TEXT" % field_name.lower()
                                query = query + s
                            query = query + ")"
                            c.execute(query)
                            l = Log(self.connection, log_name)
                            break
                    except sqlite.Error as e:
                        logging.exception(e)
                        # Data is not valid - inform the user.
                        error(parent=ln.dialog, message="Database error. Try another log name.")
            else:
                ln.dialog.destroy()
                return

        ln.dialog.destroy()

        # Update new or existing Log object.
        l.add_record(records)
        l.populate()

        if(not exists):
            self.logs.append(l)
            self.render_log(self.log_count-1)

        # Update statistics, etc.
        self.summary.update()
        self.application.toolbox.awards.count(self)

        info(parent=self.application.window, message="Imported %d QSOs into log '%s'." % (len(records), l.name))

        return

    def export_log_adif(self, widget=None):
        """ Export the log (that is currently selected) to an ADIF file. """
        # Get the index of the selected tab in the logbook.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        dialog = Gtk.FileChooserDialog("Export Log as ADIF",
                                       self.application.window,
                                       Gtk.FileChooserAction.SAVE,
                                      (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                       Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
        dialog.set_do_overwrite_confirmation(True)

        filter = Gtk.FileFilter()
        filter.set_name("All ADIF files (*.adi, *.ADI)")
        filter.add_pattern("*.adi")
        filter.add_pattern("*.ADI")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if(response == Gtk.ResponseType.OK):
            path = dialog.get_filename()
        else:
            path = None
        dialog.destroy()

        if(path is None):
            logging.debug("No file path specified.")
        else:

            # Retrieve the log's records from the database.
            try:
                records = log.records
            except sqlite.Error as e:
                logging.exception(e)
                error(parent=self.application.window, message="Could not retrieve the records from the SQL database. No records have been exported.")
                return

            # Write the records.
            adif = ADIF()
            try:
                adif.write(records, path)
                info(parent=self.application.window, message="Exported %d QSOs to %s in ADIF format." % (len(records), path))
            except IOError as e:
                error(parent=self.application.window, message="Could not export the records. I/O error %d: %s" % (e.errno, e.strerror))
            except Exception as e:  # All other exceptions.
                logging.exception(e)
                error(parent=self.application.window, message="Could not export the records.")

        return

    def export_log_cabrillo(self, widget=None):
        """ Export the log (that is currently selected) to a Cabrillo file. """
        # Get the index of the selected tab in the logbook.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        dialog = Gtk.FileChooserDialog("Export Log as Cabrillo",
                                       self.application.window,
                                       Gtk.FileChooserAction.SAVE,
                                      (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                       Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
        dialog.set_do_overwrite_confirmation(True)

        filter = Gtk.FileFilter()
        filter.set_name("All Cabrillo files (*.log, *.LOG)")
        filter.add_pattern("*.log")
        filter.add_pattern("*.LOG")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if(response == Gtk.ResponseType.OK):
            path = dialog.get_filename()
        else:
            path = None
        dialog.destroy()

        if(path is None):
            logging.debug("No file path specified.")
        else:
            # Get Cabrillo-specific fields, such as the callsign used during a contest and the contest's name.
            ced = CabrilloExportDialog(self.application)
            response = ced.dialog.run()
            if(response == Gtk.ResponseType.OK):
                contest = ced.contest
                mycall = ced.mycall
            else:
                ced.dialog.destroy()
                return
            ced.dialog.destroy()

            # Retrieve the log's records from the database.
            try:
                records = log.records
            except sqlite.Error as e:
                logging.exception(e)
                error(parent=self.application.window, message="Could not retrieve the records from the SQL database. No records have been exported.")
                return

            # Write the records.
            cabrillo = Cabrillo()
            try:
                cabrillo.write(records, path, contest=contest, mycall=mycall)
                info(parent=self.application.window, message="Exported %d QSOs to %s in Cabrillo format." % (len(records), path))
            except IOError as e:
                error(parent=self.application.window, message="Could not export the records. I/O error %d: %s" % (e.errno, e.strerror))
            except Exception as e:  # All other exceptions.
                logging.exception(e)
                error(parent=self.application.window, message="Could not export the records.")

        return

    def print_log(self, widget=None):
        """ Print all the records in the log (that is currently selected).
        Note that only a few important fields are printed because of the restricted width of the page. """

        # Get the index of the selected tab in the logbook.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        # Retrieve the records.
        try:
            records = log.records
        except sqlite.Error as e:
            logging.exception(e)
            error(parent=self.application.window, message="Could not retrieve the records from the SQL database. No records have been printed.")
            return

        # Print the records.
        printer = Printer(self.application)
        printer.print_records(records, title="Log: %s" % log.name)

        return

    def add_record_callback(self, widget):
        """ A callback function used to add a particular record/QSO. """
        # Get the index of the selected tab in the logbook.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        # Keep the dialog open after adding a record?
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])
        (section, option) = ("general", "keep_open")
        if(have_config and config.has_option(section, option)):
            keep_open = config.getboolean("general", "keep_open")
        else:
            keep_open = False

        adif = ADIF()

        exit = False
        while not exit:
            rd = RecordDialog(application=self.application, log=log, index=None)

            all_valid = False  # Are all the field entries valid?

            # Shall we exit the while loop (and therefore close the Add Record dialog)?
            if keep_open:
                exit = False
            else:
                exit = True

            while not all_valid:
                # This while loop gives the user infinite attempts at giving valid data.
                # The add/edit record window will stay open until the user gives valid data,
                # or until the Cancel button is clicked.
                all_valid = True
                response = rd.dialog.run()
                if(response == Gtk.ResponseType.OK):
                    fields_and_data = {}
                    field_names = AVAILABLE_FIELD_NAMES_ORDERED
                    for i in range(0, len(field_names)):
                        # Validate user input.
                        fields_and_data[field_names[i]] = rd.get_data(field_names[i])
                        if(not(adif.is_valid(field_names[i], fields_and_data[field_names[i]], AVAILABLE_FIELD_NAMES_TYPES[field_names[i]]))):
                            # Data is not valid - inform the user.
                            error(parent=rd.dialog, message="The data in field \"%s\" is not valid!" % field_names[i])
                            all_valid = False
                            break  # Don't check the other data until the user has fixed the current one.

                    if(all_valid):
                        # All data has been validated, so we can go ahead and add the new record.
                        try:
                            log.add_record(fields_and_data)
                        except (sqlite.Error, IndexError) as e:
                            logging.exception(e)
                            error(parent=self.application.window, message="Could not add the record to the log.")

                        # Scroll to the new record's row in the treeview (but don't select it).
                        try:
                            record_count = log.record_count
                            treepath = Gtk.TreePath(record_count-1)
                            self.treeview[log_index].scroll_to_cell(treepath)
                        except (sqlite.Error, IndexError) as e:
                            logging.exception(e)

                        # Update summary, etc.
                        self.summary.update()
                        self.application.toolbox.awards.count(self)

                else:
                    exit = True
                    break
            rd.dialog.destroy()
        return

    def delete_record_callback(self, widget):
        """ A callback function used to delete a particular record/QSO. """

        # Get the log index.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        (sort_model, path) = self.treeselection[log_index].get_selected_rows()  # Get the selected row in the log
        try:
            sort_iter = sort_model.get_iter(path[0])
            filter_iter = self.sorter[log_index].convert_iter_to_child_iter(sort_iter)
            # ...and the ListStore model (i.e. the log) is a child of the filter model.
            child_iter = self.filter[log_index].convert_iter_to_child_iter(filter_iter)
            row_index = log.get_value(child_iter, 0)
        except IndexError:
            logging.debug("Trying to delete a record, but there are no records in the log!")
            return

        response = question(parent=self.application.window, message="Are you sure you want to delete record %d?" % row_index)
        if(response == Gtk.ResponseType.YES):
            # Deletes the record with index 'row_index' from the Records list.
            # 'iter' is needed to remove the record from the ListStore itself.
            try:
                log.delete_record(row_index, iter=child_iter)
            except (sqlite.Error, IndexError) as e:
                logging.exception(e)
                error(parent=self.application.window, message="Could not delete the record from the log.")

            # Update summary, etc.
            self.summary.update()
            self.application.toolbox.awards.count(self)

        return

    def edit_record_callback(self, widget, path=None, view_column=None):
        """ A callback function used to edit a particular record/QSO.
        Note that the widget, path and view_column arguments are not used,
        but need to be passed in since they are associated with the row-activated signal
        which is generated when the user double-clicks on a record. """

        # Get the log index.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return
        log = self.logs[log_index]

        (sort_model, path) = self.treeselection[log_index].get_selected_rows()  # Get the selected row in the log.
        try:
            sort_iter = sort_model.get_iter(path[0])
            filter_iter = self.sorter[log_index].convert_iter_to_child_iter(sort_iter)
            # ...and the ListStore model (i.e. the log) is a child of the filter model.
            child_iter = self.filter[log_index].convert_iter_to_child_iter(filter_iter)
            row_index = log.get_value(child_iter, 0)
        except IndexError:
            logging.debug("Could not find the selected row's index!")
            return

        rd = RecordDialog(application=self.application, log=self.logs[log_index], index=row_index)
        all_valid = False  # Are all the field entries valid?

        adif = ADIF()
        while(not all_valid):
            # This while loop gives the user infinite attempts at giving valid data.
            # The add/edit record window will stay open until the user gives valid data,
            # or until the Cancel button is clicked.
            all_valid = True
            response = rd.dialog.run()
            if(response == Gtk.ResponseType.OK):
                fields_and_data = {}
                field_names = AVAILABLE_FIELD_NAMES_ORDERED
                for i in range(0, len(field_names)):
                    # Validate user input.
                    fields_and_data[field_names[i]] = rd.get_data(field_names[i])
                    if(not(adif.is_valid(field_names[i], fields_and_data[field_names[i]], AVAILABLE_FIELD_NAMES_TYPES[field_names[i]]))):
                        # Data is not valid - inform the user.
                        error(parent=rd.dialog, message="The data in field \"%s\" is not valid!" % field_names[i])
                        all_valid = False
                        break  # Don't check the other fields until the user has fixed the current field's data.

                if(all_valid):
                    try:
                        # Get the record in its current state from the database.
                        record = log.get_record_by_index(row_index)
                        # Iterate over all fields and check whether the data has actually changed. Database updates can be expensive.
                        for i in range(0, len(field_names)):
                            if(record[field_names[i].lower()] != fields_and_data[field_names[i]]):
                                # Update the record in the database and then in the ListStore.
                                # We add 1 onto the column_index here because we don't want to consider the index column.
                                log.edit_record(row_index, field_names[i], fields_and_data[field_names[i]], iter=child_iter, column_index=i+1)
                    except(sqlite.Error, IndexError) as e:
                        logging.exception(e)
                        error(parent=rd.dialog, message="Could not edit record %d." % row_index)

                    # Update summary, etc.
                    self.summary.update()
                    self.application.toolbox.awards.count(self)

        rd.dialog.destroy()
        return

    def remove_duplicates_callback(self, widget=None):
        """ A callback function used to remove duplicate records in a log.
        Detecting duplicate records is done based on the CALL, QSO_DATE, and TIME_ON fields. """
        logging.debug("Removing duplicate records...")

        # Get the log index.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return

        log = self.logs[log_index]

        (number_of_duplicates, number_of_duplicates_removed) = log.remove_duplicates()
        info(parent=self.application.window, message="Found %d duplicate(s). Successfully removed %d duplicate(s)." % (number_of_duplicates, number_of_duplicates_removed))

        if(number_of_duplicates_removed > 0):
            # Update statistics.
            self.summary.update()
            self.application.toolbox.awards.count(self)

        return

    def record_count_callback(self, widget=None):
        """ A callback function used to show the record count for the selected log. """

        # Get the log index.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            error(parent=self.application.window, message=e)
            return

        # Get the number of records.
        log = self.logs[log_index]
        try:
            record_count = log.record_count
            info(parent=self.application.window, message="Log '%s' contains %d records." % (log.name, record_count))
        except sqlite.Error as e:
            logging.exception(e)
            error(parent=self.application.window, message="Could not get the record count for '%s' because of a database error." % log.name)

        return

    def pinpoint_callback(self, widget=None, path=None):
        """ A callback function used to pinpoint the callsign on the world map. """

        try:
            log_index = self.get_log_index()
            row_index = self.get_record_index()
            if(log_index is None or row_index is None):
                raise ValueError("Could not determine the log and/or record index.")
            r = self.logs[log_index].get_record_by_index(row_index)
        except ValueError as e:
            logging.error(e)
            return

        self.application.toolbox.world_map.pinpoint(r)

        return

    def copy_callback(self, widget=None, path=None):
        """ A callback function used to copy selected logs. """

        try:
            log_index = self.get_log_index()
            row_index = self.get_record_index()
            if(log_index is None or row_index is None):
                raise ValueError("Could not determine the log and/or record index.")
            r = self.logs[log_index].get_record_by_index(row_index)
        except ValueError as e:
            logging.error(e)
            return

        d = {}
        for key in r.keys():
            d[key.upper()] = r[key]
        j = json.dumps(d)

        self.application.clipboard.set_text(j, len(j))

        return

    def clipboard_text_received(self, clipboard, text, log):
        r = json.loads(text)
        log.add_record(r)
        return

    def paste_callback(self, widget=None, path=None):
        """ A callback function used to paste selected logs. """

        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("Could not determine the log index.")
            l = self.logs[log_index]
        except ValueError as e:
            logging.error(e)
            return

        self.application.clipboard.request_text(self.clipboard_text_received, l)

        return

    @property
    def log_count(self):
        """ Return the total number of logs in the logbook.

        :returns: The total number of logs in the logbook.
        :rtype: int
        """
        return len(self.logs)

    @property
    def record_count(self):
        """ Return the total number of QSOs/records in the whole logbook.

        :returns: The total number of QSOs/records in the whole logbook.
        :rtype: int
        """
        return sum([log.record_count for log in self.logs])

    def log_name_exists(self, table_name):
        """ Determine whether a Log object with a given name exists in the SQL database.

        :arg str table_name: The name of the log (i.e. the name of the table in the SQL database).
        :returns: True if the log name already exists in the logbook; False if it does not already exist.
        :rtype: bool
        :raises sqlite.Error: If a database error occurs.
        """

        with self.connection:
            c = self.connection.cursor()
            c.execute("SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE name=?)", [table_name])
            exists = c.fetchone()
        if(exists[0] == 1):
            return True
        else:
            return False

    def get_log_index(self, name=None):
        """ Given the name of a log, return its index in the list of Log objects.

        :arg str name: The name of the log. If None, use the name of the currently-selected log.
        :returns: The index of the named log in the list of Log objects. Returns None if the log cannot be found.
        :rtype: int
        """
        if(name is None):
            # If no page name is supplied, then just use the currently selected page.
            page_index = self.notebook.get_current_page()  # Get the index of the selected tab in the logbook.
            if(page_index == 0 or page_index == self.notebook.get_n_pages()-1):
                # We either have the Summary page, or the "+" (add log) blank/dummy page.
                logging.debug("No log currently selected!")
                return None
            name = self.notebook.get_nth_page(page_index).get_name()
        # If a page of the logbook (and therefore a Log object) gets deleted,
        # then the page_index may not correspond to the index of the log in the self.logs list.
        # Therefore, we have to search for the tab with the same name as the log.
        log_index = None
        for i in range(0, len(self.logs)):
            if(self.logs[i].name == name):
                log_index = i
                break
        return log_index

    def get_record_index(self):
        """ Return the index of the currently selected record.

        :returns: The index of the currently selected record in the currently selected log. Returns None if the record or log cannot be found.
        :rtype: int
        """

        # Get the index of the selected log.
        try:
            log_index = self.get_log_index()
            if(log_index is None):
                raise ValueError("The log index could not be determined. Perhaps the Summary page is selected?")
        except ValueError as e:
            logging.error(e)
            return None
        log = self.logs[log_index]

        # Get the selected row in the log.
        (sort_model, path) = self.treeselection[log_index].get_selected_rows()
        try:
            sort_iter = sort_model.get_iter(path[0])
            filter_iter = self.sorter[log_index].convert_iter_to_child_iter(sort_iter)
            # ...and the ListStore model (i.e. the log) is a child of the filter model.
            child_iter = self.filter[log_index].convert_iter_to_child_iter(filter_iter)
            row_index = log.get_value(child_iter, 0)
        except IndexError:
            logging.error("Could not find the selected row's index!")
            return None

        return row_index

    def get_logs(self):
        """ Retrieve all the logs in the logbook file, and create Log objects that represent them.

        :returns: A list containing all the logs in the logbook.
        :rtype: list
        :raises sqlite.Error: If the log names could not be determined from the sqlite_master table in the database.
        """
        logs = []
        with self.connection:
            c = self.connection.cursor()
            c.execute("SELECT name FROM sqlite_master WHERE type='table' AND name NOT GLOB 'sqlite_*'")
            for name in c:
                l = Log(self.connection, name[0])
                l.populate()
                logs.append(l)
        return logs



#!/usr/bin/env python3

#    Copyright (C) 2013-2018 Christian Thomas Jacobs.

#    This file is part of PyQSO.

#    PyQSO is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyQSO is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyQSO.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject
import logging
import sqlite3 as sqlite
import re
from os.path import expanduser
from datetime import datetime
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
try:
    import numpy
    logging.info("Using version %s of numpy." % (numpy.__version__))
    import matplotlib
    logging.info("Using version %s of matplotlib." % (matplotlib.__version__))
    import cartopy
    logging.info("Using version %s of cartopy." % (cartopy.__version__))
    from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvas
    from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3
    have_necessary_modules = True
except ImportError as e:
    logging.warning(e)
    logging.warning("Could not import a non-standard Python module needed by the WorldMap class, or the version of the non-standard module is too old. Check that all the PyQSO dependencies are satisfied.")
    have_necessary_modules = False
try:
    import geocoder
    have_geocoder = True
except ImportError:
    logging.warning("Could not import the geocoder module!")
    have_geocoder = False

if(have_necessary_modules):
    class NavigationToolbar(NavigationToolbar2GTK3):
        """ Navigation tools for the World Map. """
        # Only include a subset of the tools.
        toolitems = [t for t in NavigationToolbar2GTK3.toolitems if t[0] in ("Home", "Zoom", "Save")]


class Point:
    """ A point on the grey line map. """
    def __init__(self, name, latitude, longitude, style="yo"):
        """ Set up the point's attributes.

        :arg str name: The name that identifies the point.
        :arg float latitude: The latitude of the point on the map.
        :arg float longitude: The longitude of the point on the map.
        :arg str style: The style of the point when plotted. By default it is a filled yellow circle.
        """

        self.name = name
        self.latitude = latitude
        self.longitude = longitude
        self.style = style
        return


class Maidenhead:

    """ The Maidenhead Locator System. """

    def __init__(self):
        self.upper = "ABCDEFGHIJKLMNOPQR"
        self.lower = "abcdefghijklmnopqrstuvwx"
        return

    def ll2gs(self, latitude, longitude, subsquare=False):
        """ Convert latitude-longitude coordinates to a Maidenhead grid square locator.
        This is based on the code by Walter Underwood, K6WRU (https://ham.stackexchange.com/questions/221/how-can-one-convert-from-lat-long-to-grid-square).

        :arg float latitude: The latitude.
        :arg float longitude: The longitude.
        :arg bool subsquare: Option to include the subsquare (thereby obtaining a 6-character Maidenhead locator).
        :rtype: str
        :returns: The Maidenhead grid square locator.
        """

        adjusted_latitude = latitude + 90
        adjusted_longitude = longitude + 180
        field_latitude = self.upper[int(adjusted_latitude/10)]
        field_longitude = self.upper[int(adjusted_longitude/20)]
        square_latitude = int(adjusted_latitude % 10)
        square_longitude = int((adjusted_longitude/2) % 10)

        if(subsquare):
            adjusted_latitude_remainder = (adjusted_latitude - int(adjusted_latitude)) * 60
            adjusted_longitude_remainder = ((adjusted_longitude) - int(adjusted_longitude/2)*2) * 60
            subsquare_latitude = self.lower[int(adjusted_latitude_remainder/2.5)]
            subsquare_longitude = self.lower[int(adjusted_longitude_remainder/5)]
            return ("%s"*6) % (field_longitude, field_latitude, square_longitude, square_latitude, subsquare_longitude, subsquare_latitude)
        else:
            return ("%s"*4) % (field_longitude, field_latitude, square_longitude, square_latitude)

    def gs2ll(self, grid_square):
        """ Convert a Maidenhead grid square locator to latitude-longitude coordinates.
        This is based on the gridSquareToLatLon function in HamGridSquare.js by Paul Brewer, KI6CQ (https://gist.github.com/DrPaulBrewer/4279e9d234a1bd6dd3c0), released under the MIT license.

        :arg str grid_square: The Maidenhead grid square locator.
        :rtype: tuple
        :returns: The latitude-longitude coordinates in a tuple.
        """

        m = re.match(r"^[A-X][A-X][0-9][0-9]$", grid_square)
        if(m):
            gs = m.group(0)
            latitude = self.latitude4(gs)+0.5
            longitude = self.longitude4(gs)+1.0
        else:
            m = re.match(r"^[A-X][A-X][0-9][0-9][a-x][a-x]$", grid_square)
            if(m):
                gs = m.group(0)
                latitude = self.latitude4(gs) + (1.0/60.0)*2.5*(ord(gs[5])-ord("a")+0.5)
                longitude = self.longitude4(gs) + (1.0/60.0)*5*(ord(gs[4])-ord("a")+0.5)
            else:
                raise ValueError("Unable to parse grid square string.")

        return (latitude, longitude)

    def latitude4(self, g):
        return 10*(ord(g[1]) - ord("A")) + int(g[3])-90

    def longitude4(self, g):
        return 20*(ord(g[0]) - ord("A")) + 2*int(g[2])-180


class WorldMap:

    """ A tool for visualising the world map. """

    def __init__(self, application):
        """ Set up the drawing canvas and the timer which will re-plot the world map every 30 minutes.

        :arg application: The PyQSO application containing the main Gtk window, etc.
        """
        logging.debug("Setting up the world map...")

        self.application = application
        self.builder = self.application.builder
        self.points = []

        if(have_necessary_modules):
            self.fig = matplotlib.figure.Figure()
            self.canvas = FigureCanvas(self.fig)  # For embedding in the Gtk application
            self.builder.get_object("world_map").pack_start(self.canvas, True, True, 0)
            toolbar = NavigationToolbar(self.canvas, self.application.window)
            self.builder.get_object("world_map").pack_start(toolbar, False, False, 0)
            self.refresh_event = GObject.timeout_add(1800000, self.draw)  # Re-draw the world map automatically after 30 minutes (if the world map tool is visible).

        # Add the QTH coordinates for plotting, if available.
        config = configparser.ConfigParser()
        have_config = (config.read(expanduser('~/.config/pyqso/preferences.ini')) != [])
        (section, option) = ("world_map", "show_qth")
        if(have_config and config.has_option(section, option)):
            if(config.getboolean(section, option)):
                try:
                    qth_name = config.get("world_map", "qth_name")
                    qth_latitude = float(config.get("world_map", "qth_latitude"))
                    qth_longitude = float(config.get("world_map", "qth_longitude"))
                    self.add_point(qth_name, qth_latitude, qth_longitude, "ro")
                except ValueError:
                    logging.warning("Unable to get the QTH name, latitude and/or longitude. The QTH will not be pinpointed on the world map. Check preferences?")

        # Maidenhead grid squares.
        self.maidenhead = Maidenhead()
        self.show_grid_squares = False
        self.shade_worked_grid_squares = False
        (section, option) = ("world_map", "show_grid_squares")
        if(have_config and config.has_option(section, option)):
            self.show_grid_squares = config.getboolean(section, option)
            (section, option) = ("world_map", "shade_worked_grid_squares")
            if(have_config and config.has_option(section, option)):
                self.shade_worked_grid_squares = config.getboolean(section, option)

        self.builder.get_object("world_map").show_all()

        logging.debug("World map ready!")

        return

    def add_point(self, name, latitude, longitude, style="yo"):
        """ Add a point and re-draw the map.

        :arg str name: The name that identifies the point.
        :arg float latitude: The latitude of the point on the map.
        :arg float longitude: The longitude of the point on the map.
        :arg str style: The style of the point when plotted. By default it is a filled yellow circle.
        """
        p = Point(name, latitude, longitude, style)
        self.points.append(p)
        self.draw()
        return

    def pinpoint(self, r):
        """ Pinpoint the location of a QSO on the world map.

        :arg r: The QSO record containing the location to pinpoint.
        """

        if(have_geocoder):
            callsign = r["CALL"]
            gridsquare = r["GRIDSQUARE"]
            country = r["COUNTRY"]

            # Get the latitude-longitude coordinates. Use any GRIDSQUARE information first since this is likely to be more accurate than the COUNTRY field.
            if(gridsquare):
                try:
                    latitude, longitude = self.maidenhead.gs2ll(gridsquare)
                    logging.debug("QTH coordinates found: (%s, %s)", str(latitude), str(longitude))
                    self.add_point(callsign, latitude, longitude)
                    return
                except ValueError:
                    logging.exception("Unable to lookup QTH coordinates.")

            if(country):
                try:
                    g = geocoder.google(country)
                    latitude, longitude = g.latlng
                    logging.debug("QTH coordinates found: (%s, %s)", str(latitude), str(longitude))
                    self.add_point(callsign, latitude, longitude)
                    return
                except ValueError:
                    logging.exception("Unable to lookup QTH coordinates.")
                except Exception:
                    logging.exception("Unable to lookup QTH coordinates. Check connection to the internets? Lookup limit reached?")

        return

    def get_worked_grid_squares(self, logbook):
        """ Get the array of worked grid squares.

        :arg logbook: The logbook containing logs which in turn contain QSOs.
        :returns: A two-dimensional array of boolean values showing which grid squares have been worked.
        :rtype: numpy.array
        """

        worked_grid_squares = numpy.zeros((len(self.maidenhead.upper), len(self.maidenhead.upper)), dtype=bool)

        for log in logbook.logs:
            try:
                records = log.records
                for r in records:
                    if(r["GRIDSQUARE"]):
                        grid_square = r["GRIDSQUARE"][0:2].upper()  # Only consider the field value (e.g. IO).
                        worked_grid_squares[self.maidenhead.upper.index(grid_square[1]), self.maidenhead.upper.index(grid_square[0])] = True

            except sqlite.Error as e:
                logging.error("Could not update the array of worked grid squares for log '%s' because of a database error." % log.name)
                logging.exception(e)

        return worked_grid_squares

    def draw(self):
        """ Draw the world map and the grey line on top of it.

        :returns: Always returns True to satisfy the GObject timer, unless the necessary WorldMap dependencies are not satisfied (in which case, the method returns False so as to not re-draw the canvas).
        :rtype: bool
        """

        if(have_necessary_modules):
            toolbox = self.builder.get_object("toolbox")
            tools = self.builder.get_object("tools")
            if(tools.get_current_page() != 1 or not toolbox.get_visible()):
                # Don't re-draw if the world map is not visible.
                return True  # We need to return True in case this is method was called by a timer event.
            else:
                # Set up the world map.
                logging.debug("Drawing the world map...")
                self.fig.clf()
                ax = self.fig.add_subplot(111, projection=cartopy.crs.PlateCarree())
                ax.set_extent([-180, 180, -90, 90])
                ax.set_aspect("auto")

                gl = ax.gridlines(draw_labels=True)
                gl.xlabels_top = False
                gl.ylabels_right = False
                gl.xformatter = cartopy.mpl.gridliner.LONGITUDE_FORMATTER
                gl.yformatter = cartopy.mpl.gridliner.LATITUDE_FORMATTER
                ax.add_feature(cartopy.feature.LAND, facecolor="olivedrab")
                ax.add_feature(cartopy.feature.OCEAN, facecolor="cornflowerblue")
                ax.add_feature(cartopy.feature.COASTLINE)
                ax.add_feature(cartopy.feature.BORDERS, alpha=0.4)

                # Draw the grey line. This is based on the code from the Cartopy Aurora Forecast example (http://scitools.org.uk/cartopy/docs/latest/gallery/aurora_forecast.html) and used under the Open Government Licence (http://scitools.org.uk/cartopy/docs/v0.15/copyright.html).
                logging.debug("Drawing the grey line...")
                dt = datetime.utcnow()
                axial_tilt = 23.5
                reference_solstice = datetime(2016, 6, 21, 22, 22)
                days_per_year = 365.2425
                seconds_per_day = 86400.0

                days_since_reference = (dt - reference_solstice).total_seconds()/seconds_per_day
                latitude = axial_tilt*numpy.cos(2*numpy.pi*days_since_reference/days_per_year)
                seconds_since_midnight = (dt - datetime(dt.year, dt.month, dt.day)).seconds
                longitude = -(seconds_since_midnight/seconds_per_day - 0.5)*360

                pole_longitude = longitude
                if latitude > 0:
                    pole_latitude = -90 + latitude
                    central_rotated_longitude = 180
                else:
                    pole_latitude = 90 + latitude
                    central_rotated_longitude = 0

                rotated_pole = cartopy.crs.RotatedPole(pole_latitude=pole_latitude, pole_longitude=pole_longitude, central_rotated_longitude=central_rotated_longitude)

                x = numpy.empty(360)
                y = numpy.empty(360)
                x[:180] = -90
                y[:180] = numpy.arange(-90, 90.)
                x[180:] = 90
                y[180:] = numpy.arange(90, -90., -1)

                ax.fill(x, y, transform=rotated_pole, color="black", alpha=0.5)

                # Plot points on the map.
                if(self.points):
                    logging.debug("Plotting QTHs on the map...")
                    for p in self.points:
                        ax.plot(p.longitude, p.latitude, p.style, transform=cartopy.crs.PlateCarree())
                        projected_x, projected_y = ax.projection.transform_point(p.longitude, p.latitude, src_crs=cartopy.crs.PlateCarree())
                        ax.annotate(p.name, xy=(projected_x, projected_y), xytext=(0, 2.5), textcoords="offset points", color="white", size="small", weight="bold")

                # Draw Maidenhead grid squares and shade in the worked squares.
                x = numpy.linspace(-180, 180, len(list(self.maidenhead.upper))+1)
                y = numpy.linspace(-90, 90, len(list(self.maidenhead.upper))+1)
                if(self.show_grid_squares):
                    if(self.shade_worked_grid_squares):
                        worked_grid_squares = self.get_worked_grid_squares(self.application.logbook)
                        masked = numpy.ma.masked_array(worked_grid_squares, worked_grid_squares == 0)
                    else:
                        z = numpy.zeros((len(self.maidenhead.upper), len(self.maidenhead.upper)), dtype=bool)
                        masked = numpy.ma.masked_array(z, z == 0)
                    ax.pcolormesh(x, y, masked, transform=cartopy.crs.PlateCarree(), cmap="Reds", vmin=0, vmax=1, edgecolors="k", linewidth=1.5, alpha=0.4)

                    # Grid square labels.
                    for i in range(len(self.maidenhead.upper)):
                        for j in range(len(self.maidenhead.upper)):
                            text = self.maidenhead.upper[i]+self.maidenhead.upper[j]
                            ax.text((x[i]+x[i+1])/2.0, (y[j]+y[j+1])/2.0, text, ha="center", va="center", size="small", color="w", family="monospace", alpha=0.4)

                return True
        else:
            return False  # Don't try to re-draw the canvas if the necessary modules to do so could not be imported.
